from copy import copy
from operator import eq

from hwt.doc_markers import internal
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.types.bitValFunctions import bitsCmp, \
    bitsBitOp, bitsArithOp
from hwt.hdl.types.bitVal_opReduce import tryReduceOr, tryReduceAnd, \
    tryReduceXor
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import BOOL, INT, BIT, SLICE, BIT_N
from hwt.hdl.types.eventCapableVal import EventCapableVal
from hwt.hdl.types.slice import HSlice
from hwt.hdl.types.sliceUtils import slice_to_SLICE
from hwt.hdl.types.typeCast import toHVal
from hwt.hdl.value import HValue, areHValues
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import SignalDriverErr
from pyMathBitPrecise.bits3t import Bits3val
from pyMathBitPrecise.bits3t_vld_masks import vld_mask_for_xor, vld_mask_for_and, \
    vld_mask_for_or
from pyMathBitPrecise.bit_utils import ValidityError


class BitsVal(Bits3val, EventCapableVal, HValue):
    """
    :attention: operator on signals are using value operator functions as well
    """
    _BOOL = Bits(1, name="bool")

    @classmethod
    def from_py(cls, typeObj, val, vld_mask=None):
        val, vld_mask = typeObj._normalize_val_and_mask(val, vld_mask)
        return cls(typeObj, val, vld_mask=vld_mask)

    @internal
    def _convSign__val(self, signed):
        v = Bits3val.cast_sign(self, signed)
        if signed is not None:
            assert v._dtype is not self._dtype, "can modify shared type instance"
            v._dtype.force_vector = True
        return v

    @internal
    def _convSign(self, signed):
        """
        Convert signum, no bit manipulation just data are represented
        differently

        :param signed: if True value will be signed,
            if False value will be unsigned,
            if None value will be vector without any sign specification
        """
        if isinstance(self, HValue):
            return self._convSign__val(signed)
        else:
            if self._dtype.signed == signed:
                return self
            t = copy(self._dtype)
            t.signed = signed
            if t.signed is not None:
                t.force_vector = True

            if signed is None:
                cnv = AllOps.BitsAsVec
            elif signed:
                cnv = AllOps.BitsAsSigned
            else:
                cnv = AllOps.BitsAsUnsigned

            return Operator.withRes(cnv, [self], t)

    def _auto_cast(self, dtype):
        return HValue._auto_cast(self, dtype)

    def _signed(self):
        return self._convSign(True)

    def _unsigned(self):
        return self._convSign(False)

    def _vec(self):
        return self._convSign(None)

    @internal
    def _concat__val(self, other):
        return Bits3val._concat(self, other)

    def _concat(self, other):
        """
        Concatenate this with other to one wider value/signal
        """
        w = self._dtype.bit_length()
        try:
            other._dtype.bit_length
        except AttributeError:
            raise TypeError("Can not concat Bits and", other)

        self = self._vec()
        if areHValues(self, other):
            return self._concat__val(other)
        else:
            w = self._dtype.bit_length()
            other_w = other._dtype.bit_length()
            resWidth = w + other_w
            Bits = self._dtype.__class__
            resT = Bits(resWidth, signed=self._dtype.signed, force_vector=True)
            # is instance of signal
            if isinstance(other, InterfaceBase):
                other = other._sig

            if other._dtype == BOOL:
                other = other._auto_cast(BIT)
            elif isinstance(other._dtype, Bits):
                if other._dtype.signed is not None:
                    other = other._vec()
            else:
                raise TypeError(other._dtype)

            if self._dtype.signed is not None:
                self = self._vec()

            return Operator.withRes(AllOps.CONCAT, [self, other], resT)\
                           ._auto_cast(Bits(resWidth,
                                            signed=self._dtype.signed))

    def __getitem__(self, key):
        """
        [] operator

        :attention: Table below is for litle endian bit order (MSB:LSB)
            which is default.
            This is **reversed** as it is in pure python
            where it is [0, len(self)].

        :attention: slice on slice f signal is automatically reduced
            to single slice

        +-----------------------------+----------------------------------------------------------------------------------+
        | a[up:low]                   | items low through up; a[16:8] selects upper byte from 16b vector a               |
        +-----------------------------+----------------------------------------------------------------------------------+
        | a[up:]                      | low is automatically substituted with 0; a[8:] will select lower 8 bits          |
        +-----------------------------+----------------------------------------------------------------------------------+
        | a[:end]                     | up is automatically substituted; a[:8] will select upper byte from 16b vector a  |
        +-----------------------------+----------------------------------------------------------------------------------+
        | a[:], a[-1], a[-2:], a[:-2] | raises NotImplementedError   (not implemented due to complicated support in hdl) |
        +-----------+----------------------------------------------------------------------------------------------------+
        """
        st = self._dtype
        length = st.bit_length()
        if length == 1 and not st.force_vector:
            # assert not indexing on single bit
            raise TypeError("indexing on single bit")

        if isinstance(key, slice):
            key = slice_to_SLICE(key, length)
            isSLICE = True
        else:
            isSLICE = isinstance(key, HSlice.getValueCls())

        if isSLICE:
            # :note: downto notation
            start = key.val.start
            stop = key.val.stop
            if key.val.step != -1:
                raise NotImplementedError()
            startIsVal = isinstance(start, HValue)
            stopIsVal = isinstance(stop, HValue)
            indexesareHValues = startIsVal and stopIsVal
        else:
            key = toHVal(key, INT)

        iamVal = isinstance(self, HValue)
        iAmResultOfIndexing = (not iamVal and
                               len(self.drivers) == 1 and
                               isinstance(self.origin, Operator) and
                               self.origin.operator == AllOps.INDEX)

        Bits = self._dtype.__class__
        if isSLICE:
            if indexesareHValues and start.val == length and stop.val == 0:
                # selecting all bits no conversion needed
                return self

            if iAmResultOfIndexing:
                # try reduce self and parent slice to one
                original, parentIndex = self.origin.operands
                if isinstance(parentIndex._dtype, HSlice):
                    parentLower = parentIndex.val.stop
                    start = start + parentLower
                    stop = stop + parentLower
                    return original[start:stop]

            # check start boundaries
            if startIsVal:
                _start = int(start)
                if _start < 0 or _start > length:
                    raise IndexError(_start, length)

            # check end boundaries
            if stopIsVal:
                _stop = int(stop)
                if _stop < 0 or _stop > length:
                    raise IndexError(_stop, length)

            # check width of selected range
            if startIsVal and stopIsVal and _start - _stop <= 0:
                raise IndexError(_start, _stop)

            if iamVal:
                if isinstance(key, SLICE.getValueCls()):
                    key = key.val
                v = Bits3val.__getitem__(self, key)
                if v._dtype.bit_length() == 1 and not v._dtype.force_vector:
                    assert v._dtype is not self._dtype
                    v._dtype.force_vector = True
                return v
            else:
                key = SLICE.from_py(slice(start, stop, -1))
                _resWidth = start - stop
                resT = Bits(bit_length=_resWidth, force_vector=True,
                            signed=st.signed, negated=st.negated)

        elif isinstance(key, Bits.getValueCls()):
            if key._is_full_valid():
                # check index range
                _v = int(key)
                if _v < 0 or _v > length - 1:
                    raise IndexError(_v)
                if iAmResultOfIndexing:
                    original, parentIndex = self.origin.operands
                    if isinstance(parentIndex._dtype, HSlice):
                        parentLower = parentIndex.val.stop
                        return original[parentLower + _v]

            if iamVal:
                return Bits3val.__getitem__(self, key)

            resT = BIT
        elif isinstance(key, RtlSignalBase):
            t = key._dtype
            if isinstance(t, HSlice):
                resT = Bits(bit_length=key.staticEval()._size(),
                            force_vector=True,
                            signed=st.signed,
                            negated=st.negated)
            elif isinstance(t, Bits):
                resT = BIT
            else:
                raise TypeError(
                    "Index operation not implemented"
                    f" for index of type {t}")

        else:
            raise TypeError(
                f"Index operation not implemented for index {key}")

        if st.negated and resT is BIT:
            resT = BIT_N
        return Operator.withRes(AllOps.INDEX, [self, key], resT)

    def __setitem__(self, index, value):
        """
        this []= operator can not be called in desing description, it can be only used to update HValues
        """
        if not isinstance(self, HValue):
            raise TypeError("To assign a member of hdl arrray/vector/list/... use a[index](val) instead of a[index] = val")

        # convert index to HSlice or hInt
        if isinstance(index, HValue):
            index = index
        elif isinstance(index, slice):
            length = self._dtype.bit_length()
            index = slice_to_SLICE(index, length)
            if not index._is_full_valid():
                raise ValueError("invalid index", index)
        else:
            index = INT.from_py(index)

        # convert value to bits of length specified by index
        if index._dtype == SLICE:
            Bits = self._dtype.__class__
            itemT = Bits(index._size())
        else:
            itemT = BIT

        if isinstance(value, HValue):
            value = value._auto_cast(itemT)
        else:
            value = itemT.from_py(value)

        return Bits3val.__setitem__(self, index, value)

    def __invert__(self):
        if isinstance(self, HValue):
            return Bits3val.__invert__(self)
        else:
            try:
                # double negation
                d = self.singleDriver()
                if isinstance(d, Operator) and d.operator == AllOps.NOT:
                    return d.operands[0]
            except SignalDriverErr:
                pass
            return Operator.withRes(AllOps.NOT, [self], self._dtype)

    def __hash__(self):
        if isinstance(self, RtlSignalBase):
            return hash(id(self))
        else:
            return Bits3val.__hash__(self)

    # comparisons
    def _eq(self, other):
        return bitsCmp(self, other, AllOps.EQ, BIT.from_py(1), eq)

    def __ne__(self, other):
        return bitsCmp(self, other, AllOps.NE, BIT.from_py(0))

    def __lt__(self, other):
        return bitsCmp(self, other, AllOps.LT, BIT.from_py(0))

    def __gt__(self, other):
        return bitsCmp(self, other, AllOps.GT, BIT.from_py(0))

    def __ge__(self, other):
        return bitsCmp(self, other, AllOps.GE, BIT.from_py(1))

    def __le__(self, other):
        return bitsCmp(self, other, AllOps.LE, BIT.from_py(1))

    def __xor__(self, other):
        return bitsBitOp(self, other, AllOps.XOR,
                         vld_mask_for_xor, tryReduceXor, self._dtype.from_py(0))

    def __and__(self, other):
        return bitsBitOp(self, other, AllOps.AND,
                         vld_mask_for_and, tryReduceAnd, self)

    def __or__(self, other):
        return bitsBitOp(self, other, AllOps.OR,
                         vld_mask_for_or, tryReduceOr, self)

    def __lshift__(self, other):
        """
        shift left

        :note: arithmetic sift if type is signed else logical shift
        """
        width = self._dtype.bit_length()
        if self._dtype.signed:
            raise NotImplementedError()

        return self[(width - int(other)):]._concat(Bits(int(other)).from_py(0))

    def __rshift__(self, other):
        """
        shift right

        :note: arithmetic sift if type is signed else logical shift
        """
        if self._dtype.signed:
            return self[:self._dtype.bit_length() - other]._concat(self[:other])

        return Bits(int(other)).from_py(0)._concat(self[:other])

    def __neg__(self):
        if isinstance(self, HValue):
            return Bits3val.__neg__(self)
        else:
            if not self._dtype.signed:
                self = self._signed()

            resT = self._dtype

            o = Operator.withRes(AllOps.MINUS_UNARY, [self], self._dtype)
            return o._auto_cast(resT)

    def __sub__(self, other):
        return bitsArithOp(self, other, AllOps.SUB)

    def __add__(self, other):
        return bitsArithOp(self, other, AllOps.ADD)

    def __floordiv__(self, other) -> "Bits3val":
        other = toHVal(other, suggestedType=self._dtype)
        if isinstance(self, HValue) and isinstance(other, HValue):
            return Bits3val.__floordiv__(self, other)
        else:
            if not isinstance(other._dtype, self._dtype.__class__):
                raise TypeError()
            return Operator.withRes(AllOps.DIV,
                                    [self, other],
                                    self._dtype.__copy__())
    def __pow__(self, other):
        raise TypeError(f"** operator not implemented for instance of {self.__class__}")

    def __mod__(self, other):
        raise TypeError(f"% operator not implemented for instance of {self.__class__}")

    def _ternary(self, a, b):
        if isinstance(self, HValue):
            if self:
                return a
            else:
                return b
        else:
            a = toHVal(a)
            b = toHVal(b, suggestedType=a._dtype)
            try:
                if a == b:
                    return a
            except ValidityError:
                pass

            if not (a._dtype == b._dtype):
                # all case values of ternary has to have same type
                b = b._auto_cast(a._dtype)

            return Operator.withRes(
                AllOps.TERNARY,
                [self, a, b],
                a._dtype.__copy__())

    def __mul__(self, other):
        Bits = self._dtype.__class__
        other = toHVal(other)
        if not isinstance(other._dtype, Bits):
            raise TypeError(other)

        self_is_val = isinstance(self, HValue)
        other_is_val = isinstance(other, HValue)

        if self_is_val and other_is_val:
            return Bits3val.__mul__(self, other)
        else:
            # reduce *1 and *0
            if self_is_val and self._is_full_valid():
                _s = int(self)
                if _s == 0:
                    return self._dtype.from_py(0)
                elif _s:
                    return other._auto_cast(self._dtype)

            if other_is_val and other._is_full_valid():
                _o = int(other)
                if _o == 0:
                    return self._dtype.from_py(0)
                elif _o == 1:
                    return self

            myT = self._dtype
            if self._dtype.signed is None:
                self = self._unsigned()

            if isinstance(other._dtype, Bits):
                s = other._dtype.signed
                if s is None:
                    other = other._unsigned()
            else:
                raise TypeError(f"{self} {AllOps.MUL} {other}")

            if other._dtype == INT:
                res_w = myT.bit_length()
                res_sign = self._dtype.signed
                subResT = resT = myT
            else:
                res_w = max(myT.bit_length(), other._dtype.bit_length())
                res_sign = self._dtype.signed or other._dtype.signed
                subResT = Bits(res_w, signed=res_sign)
                resT = Bits(res_w, signed=myT.signed)

            o = Operator.withRes(AllOps.MUL, [self, other], subResT)
            return o._auto_cast(resT)



"""
This package contains classes for representation of HDL
languages (hdl types, RtlSignal, HdlAssignmentContainer, HdlStmCodeBlockContainer etc.).

It also contains classes for representation of complex HW structures
like dense transaction template (TransTmpl, FrameTmpl).
"""
# [TODO] derive from hdlConvertorAst.hdlAst


from copy import copy
from typing import Optional

from hdlConvertorAst.hdlAst._defs import HdlIdDef
from hdlConvertorAst.hdlAst._expr import HdlValueId, HdlOp, HdlOpType
from hdlConvertorAst.hdlAst._structural import HdlModuleDec, HdlCompInst
from hdlConvertorAst.to.systemc.keywords import SYSTEMC_KEYWORDS
from hdlConvertorAst.translate.common.name_scope import LanguageKeyword, NameScope
from hwt.hdl.portItem import HdlPortItem
from hwt.interfaces.std import Clk
from hwt.serializer.generic.to_hdl_ast import ToHdlAst,\
    HWT_TO_HDLCONVEROTR_DIRECTION
from hwt.serializer.simModel.serializer import ToHdlAstSimModel
from hwt.serializer.systemC.expr import ToHdlAstSystemC_expr
from hwt.serializer.systemC.statements import ToHdlAstSystemC_statements
from hwt.serializer.systemC.type import ToHdlAstSystemC_type
from ipCorePackager.constants import DIRECTION


class ToHdlAstSystemC(ToHdlAstSystemC_expr, ToHdlAstSystemC_type,
                      ToHdlAstSystemC_statements,
                      ToHdlAst):
    """
    Serialized used to convert HWT design to SystemC code
    """
    _keywords_dict = {kw: LanguageKeyword() for kw in SYSTEMC_KEYWORDS}
    sc_in_clk = HdlValueId("sc_in_clk", obj=LanguageKeyword())
    sc_out_clk = HdlValueId("sc_out_clk", obj=LanguageKeyword())
    sc_inout_clk = HdlValueId("sc_inout_clk", obj=LanguageKeyword())
    sc_in = HdlValueId("sc_in", obj=LanguageKeyword())
    sc_out = HdlValueId("sc_out", obj=LanguageKeyword())
    sc_inout = HdlValueId("sc_inout", obj=LanguageKeyword())

    def __init__(self, name_scope: Optional[NameScope]=None):
        ToHdlAst.__init__(self, name_scope=name_scope)
        self._is_target = False
        self._in_sensitivity_list = False
        self.signalType = None

    def as_hdl_HdlModuleDec(self, o: HdlModuleDec):
        return ToHdlAstSimModel.as_hdl_HdlModuleDec(self, o)

    def as_hdl_HdlPortItem(self, o: HdlPortItem):
        i = o.getInternSig()._interface
        d = o.direction
        if isinstance(i, Clk):
            assert i._dtype.bit_length() == 1, i
            if d == DIRECTION.IN:
                t = self.sc_in_clk
            elif d == DIRECTION.OUT:
                t = self.sc_out_clk
            elif d == DIRECTION.INOUT:
                t = self.sc_inout_clk
            else:
                raise ValueError(d)
        else:
            if d == DIRECTION.IN:
                pt = self.sc_in
            elif d == DIRECTION.OUT:
                pt = self.sc_out
            elif d == DIRECTION.INOUT:
                pt = self.sc_inout
            else:
                raise ValueError(d)
            t = self.as_hdl_HdlType(o._dtype)
            t = HdlOp(HdlOpType.PARAMETRIZATION, [pt, t])

        var = HdlIdDef()
        var.direction = HWT_TO_HDLCONVEROTR_DIRECTION[o.direction]
        s = o.getInternSig()
        var.name = s.name
        var.origin = o
        var.type = t
        return var

    def as_hdl_HdlCompInst(self, o: HdlCompInst) -> HdlCompInst:
        new_o = copy(o)
        new_o.param_map = []

        orig_is_target = self._is_target
        try:
            self._is_target = True
            port_map = []
            for pi in o.port_map:
                pm = self.as_hdl_PortConnection(pi)
                port_map.append(pm)
            new_o.port_map = port_map
        finally:
            self._is_target = orig_is_target

        return new_o



import os
from random import Random
from typing import Optional
import unittest

from hwt.simulator.agentConnector import autoAddAgents, \
    collect_processes_from_sim_agents
from hwt.simulator.rtlSimulatorVcd import BasicRtlSimulatorVcd
from hwt.simulator.utils import reconnectUnitSignalsToModel, valToInt,\
    allValuesToInts
from hwt.synthesizer.dummyPlatform import DummyPlatform
from hwt.synthesizer.unit import Unit
from hwtSimApi.constants import CLK_PERIOD
from hwtSimApi.hdlSimulator import HdlSimulator
from hwtSimApi.triggers import Timer


class DummySimPlatform(DummyPlatform):
    """
    DummyPlatform which ignores the constraints
    (hardware constranints which specifying something for circuit synthesis for a vendor tool)
    """


_UNSPECIFIED = object()


class SimTestCase(unittest.TestCase):
    """
    This is TestCase class contains methods which are usually used during
    hdl simulation.

    :attention: self.procs has to be specified before runSim()
    :cvar _defaultSeed: default seed for random generator
    :cvar rtl_simulator_cls: class for RTL simulator to use
        (constructed in compileSim())
    :ivar ~.u: instance of current :class:`hwt.synthesizer.unit.Unit` for test, created in restartSim()
    :ivar ~.rtl_simulator: RTL simulator used for simulation of unit,
        created in restartSim()
    :ivar ~.hdl_simulator: the simulator which manages the communication
        between Python code and rtl_simulator instance
    :ivar ~.procs: list of simulation processes (Python generator instances),
        created in restartSim()
    :ivar ~.DEFAULT_BUILD_DIR: default directory where files for simulation should be stored
    :ivar ~.DEFAULT_LOG_DIR: default directory where simulation outputs should be stored
    :ivar ~.DEFAULT_SIMULATOR: default RTL simulator generator used on background of the test
    :ivar ~.RECOMPILE: if False the compilation of the simulation is dissabled.
        This is usefull while debugging of the simulation because compilation of simulation
        may take significant amount of time and may not be required.
    """
    # value chosen because in this position bits are changing frequently
    _defaultSeed = 317
    RECOMPILE = True
    rtl_simulator_cls = None
    hdl_simulator = None
    DEFAULT_BUILD_DIR = None  # "tmp"
    DEFAULT_LOG_DIR = "tmp"
    DEFAULT_SIMULATOR = BasicRtlSimulatorVcd

    def assertValEqual(self, first, second, msg=None):
        try:
            first = first.read()
        except AttributeError:
            pass

        if not isinstance(first, int) and first is not None:
            first = valToInt(first)

        return unittest.TestCase.assertEqual(self, first, second, msg=msg)

    def assertEmpty(self, val, msg=None):
        return unittest.TestCase.assertEqual(self, len(val), 0, msg=msg)

    def assertValSequenceEqual(self, seq1, seq2, msg=None, seq_type=None):
        """
        An equality assertion for ordered sequences (like lists and tuples).
        For the purposes of this function, a valid ordered sequence type is one
        which can be indexed, has a length, and has an equality operator.

        Args:

        :param seq1: can contain instance of values or nested list of them
        :param seq2: items are not converted, if item is None it is not checked
        :param seq_type: The expected data type of the sequences, or None if no
            data type should be enforced.
        :param msg: Optional message to use on failure instead of a list of
            differences.
        """
        seq1 = allValuesToInts(seq1)
        if len(seq1) == len(seq2):
            _seq2 = []
            # replace None in seq2 with values from seq1
            for v1, v2 in zip(seq1, seq2):
                if v2 is None:
                    v2 = v1
                _seq2.append(v2)
            seq2 = _seq2
        self.assertSequenceEqual(seq1, seq2, msg, seq_type)

    def getTestName(self):
        className, testName = self.id().split(".")[-2:]
        return f"{className:s}_{testName:s}"

    def runSim(self, until: float, name=None):
        """
        Collect sim. processes from iterface agents and run simulation
        """
        if name is None:
            if self.DEFAULT_LOG_DIR is None:
                outputFileName = None
            else:
                outputFileName = os.path.join(self.DEFAULT_LOG_DIR,
                                              self.getTestName() + ".vcd")
        else:
            outputFileName = name

        if outputFileName is not None:
            d = os.path.dirname(outputFileName)
            if d:
                os.makedirs(d, exist_ok=True)

            self.rtl_simulator.set_trace_file(outputFileName, -1)
        procs = collect_processes_from_sim_agents(self.u)
        # run simulation, stimul processes are register after initial
        # initialization
        self.hdl_simulator.run(until=until, extraProcesses=self.procs + procs)
        self.rtl_simulator.finalize()
        return self.hdl_simulator

    def randomize(self, intf):
        """
        Randomly disable and enable interface for testing purposes
        """
        assert intf._isExtern, intf
        assert intf._ag is not None, intf
        randomEnProc = simpleRandomizationProcess(self, intf._ag)
        self.procs.append(randomEnProc())

    def restartSim(self):
        """
        Set simulator to initial state and connect it to

        :return: tuple (fully loaded unit with connected simulator,
            connected simulator,
            simulation processes
            )
        """
        rtl_simulator = self.rtl_simulator_cls()
        hdl_simulator = HdlSimulator(rtl_simulator)

        unit = self.u
        reconnectUnitSignalsToModel(unit, rtl_simulator)
        autoAddAgents(unit, hdl_simulator)
        self.procs = []
        self.u, self.rtl_simulator, self.hdl_simulator = \
            unit, rtl_simulator, hdl_simulator

        return unit, rtl_simulator, self.procs

    @classmethod
    def get_unique_name(cls, unit: Unit):
        uniq_name = unit._getDefaultName()
        return f"{cls.__name__:s}__{uniq_name:s}"

    @classmethod
    def compileSim(cls, unit, build_dir: Optional[str]=_UNSPECIFIED,
                   unique_name: Optional[str]=None, onAfterToRtl=None,
                   target_platform=DummySimPlatform()):
        """
        Create simulation model and connect it with interfaces of original unit
        and decorate it with agents

        :param unit: interface level unit which you wont prepare for simulation
        :param target_platform: target platform for this synthesis
        :param build_dir: folder to where to put sim model files,
            if None temporary folder is used and then deleted
            (or simulator will be constructed in memory if possible)
        :param unique_name: name which is used as name of the module for simulation
            (if is None it is automatically generated)
        :param onAfterToRtl: callback fn(unit) which will be called
            after unit will be synthesised to RTL
            and before :class:`hwt.synthesizer.unit.Unit` instance signals are replaced
            with simulator specific ones
        """
        if build_dir == _UNSPECIFIED:
            build_dir = cls.DEFAULT_BUILD_DIR

        if unique_name is None:
            unique_name = cls.get_unique_name(unit)

        cls.rtl_simulator_cls = cls.DEFAULT_SIMULATOR.build(
            unit,
            unique_name=unique_name,
            build_dir=build_dir,
            target_platform=target_platform,
            do_compile=cls.RECOMPILE)

        if onAfterToRtl:
            onAfterToRtl(unit)

        cls.u = unit

    def compileSimAndStart(
            self,
            unit: Unit,
            build_dir: Optional[str]=_UNSPECIFIED,
            unique_name: Optional[str]=None,
            onAfterToRtl=None,
            target_platform=DummySimPlatform()):
        """
        Use this method if you did not used compileSim()
        to setup the simulator and DUT
        """
        if unique_name is None:
            t_name = self.getTestName()
            u_name = unit._getDefaultName()
            unique_name = f"{t_name:s}__{u_name:s}"
        self.compileSim(unit, build_dir, unique_name,
                        onAfterToRtl, target_platform)
        self.u = unit
        SimTestCase.setUp(self)
        return self.u

    def setUp(self):
        self._rand = Random(self._defaultSeed)
        if self.rtl_simulator_cls is not None:
            # if the simulator is not compiled it is expected
            # that it will be compiled in the test and this function
            # will be called later
            self.restartSim()


def simpleRandomizationProcess(tc: SimTestCase, agent, timeQuantum=CLK_PERIOD):
    """
    A process for simulator which will randomly enable/dissable the egent for an interface
    """
    seed = tc._rand.getrandbits(64)
    random = Random(seed)

    def randomEnProc():
        # small space at start to modify agents when they are inactive
        yield Timer(timeQuantum // 4)
        while True:
            en = random.random() < 0.5
            if agent.getEnable() != en:
                agent.setEnable(en)
            delay = int(random.random() * 2) * timeQuantum
            yield Timer(delay)

    return randomEnProc


from itertools import islice

from hwt.doc_markers import internal
from hwt.hdl.statements.assignmentContainer import HdlAssignmentContainer
from hwt.hdl.statements.codeBlockContainer import HdlStmCodeBlockContainer
from hwt.pyUtils.arrayQuery import areSetsIntersets, groupedby
from hwt.serializer.utils import HdlStatement_sort_key
from hwt.hdl.statements.utils.reduction import HdlStatement_merge_statement_lists,\
    is_mergable_statement_list


class HwtStmIncompatibleStructure(Exception):
    """
    Statements are not comparable due incompatible structure
    """


@internal
def checkIfIsTooSimple(proc):
    """check if process is just unconditional assignments
       and it is useless to merge them"""
    try:
        a, = proc.statements
        if isinstance(a, HdlAssignmentContainer):
            return True
    except ValueError:
        pass
    return False


@internal
def tryToMerge(procA: HdlStmCodeBlockContainer, procB: HdlStmCodeBlockContainer):
    """
    Try merge procB into procA

    :raise IncompatibleStructure: if merge is not possible
    :attention: procA is now result if merge has succeed
    :return: procA which is now result of merge
    """
    if (checkIfIsTooSimple(procA) or
            checkIfIsTooSimple(procB) or
            areSetsIntersets(procA._outputs, procB._sensitivity) or
            areSetsIntersets(procB._outputs, procA._sensitivity) or
            not is_mergable_statement_list(procA.statements, procB.statements)):
        raise HwtStmIncompatibleStructure()

    procA.statements = HdlStatement_merge_statement_lists(
        procA.statements, procB.statements)

    procA._outputs.extend(procB._outputs)
    procA._inputs.extend(procB._inputs)
    procA._sensitivity.extend(procB._sensitivity)

    return procA


@internal
def reduceProcesses(processes):
    """
    Try to merge processes as much is possible

    :param processes: list of processes instances
    """
    # sort to make order of merging same deterministic
    processes.sort(key=HdlStatement_sort_key, reverse=True)
    # now try to reduce processes with nearly same structure of statements into one
    # to minimize number of processes
    for _, procs in groupedby(processes, lambda p: p.rank):
        for iA, pA in enumerate(procs):
            if pA is None:
                continue
            for iB, pB in enumerate(islice(procs, iA + 1, None)):
                if pB is None:
                    continue

                try:
                    pA = tryToMerge(pA, pB)
                except HwtStmIncompatibleStructure:
                    continue
                procs[iA + 1 + iB] = None
                # procs[iA] = pA

        for p in procs:
            if p is not None:
                yield p



from hwt.doc_markers import internal
from hwt.hdl.types.hdlType import HdlType
from hwt.serializer.generic.indent import getIndent
from pyMathBitPrecise.bits3t import Bits3t

BITS_DEFAUTL_SIGNED = None
BITS_DEFAUTL_FORCEVECTOR = False
BITS_DEFAUTL_NEGATED = False


class Bits(HdlType, Bits3t):
    """
    Elemental HDL type representing bits (vector or single bit)
    """

    def __init__(self, bit_length, signed=BITS_DEFAUTL_SIGNED,
                 force_vector=BITS_DEFAUTL_FORCEVECTOR,
                 negated=BITS_DEFAUTL_NEGATED,
                 name=None,
                 const=False,
                 strict_sign=True, strict_width=True):
        """
        :param negated: if true the value is in negated form
        """
        self.negated = negated
        HdlType.__init__(self, const=const)
        bit_length = int(bit_length)
        assert bit_length > 0, bit_length
        Bits3t.__init__(self, bit_length, signed, name=name,
                        force_vector=force_vector or signed is not None,
                        strict_sign=strict_sign, strict_width=strict_width)

    @internal
    def domain_size(self):
        """
        :return: how many values can have specified type
        """
        return int(2 ** self.bit_length())

    @internal
    @classmethod
    def get_auto_cast_fn(cls):
        from hwt.hdl.types.bitsCast import convertBits
        return convertBits

    @internal
    @classmethod
    def get_reinterpret_cast_fn(cls):
        from hwt.hdl.types.bitsCast import reinterpretBits
        return reinterpretBits

    @internal
    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdl.types.bitsVal import BitsVal
            cls._valCls = BitsVal
            return cls._valCls

    def __hash__(self):
        return hash((Bits3t.__hash__(self), self.const))

    def __eq__(self, other):
        return  Bits3t.__eq__(self, other) and self.const == other.const

    def __repr__(self, indent=0, withAddr=None, expandStructs=False):
        """
        :param indent: number of indentation
        :param withAddr: if is not None is used as a additional
            information about on which address this type is stored
            (used only by HStruct)
        :param expandStructs: expand HStructTypes (used by HStruct and HArray)
        """
        constr = []
        if self.name is not None:
            constr.append('"%s"' % self.name)
        c = self.bit_length()
        if c == 1:
            constr.append("1bit")
            if self.force_vector:
                constr.append("force_vector")
        else:
            constr.append(f"{c:d}bits")

        if self.const:
            constr.append("const")

        if self.signed:
            constr.append("signed")
        elif self.signed is False:
            constr.append("unsigned")

        if not self.strict_sign:
            constr.append("strict_sign=False")

        if not self.strict_width:
            constr.append("strict_width=False")

        return "%s<%s, %s>" % (getIndent(indent),
                               self.__class__.__name__,
                               ", ".join(constr))



from hwt.synthesizer.exceptions import IntfLvlConfErr
from hwt.synthesizer.interfaceLevel.mainBases import UnitBase


def getClk(unit: UnitBase):
    """
    Get clock signal from unit instance
    """
    try:
        return unit.clk
    except AttributeError:
        pass

    raise IntfLvlConfErr(f"Can not find clock signal on unit {unit}")


def getRst(unit: UnitBase):
    """
    Get reset signal from unit instance
    """
    try:
        return unit.rst
    except AttributeError:
        pass

    try:
        return unit.rst_n
    except AttributeError:
        pass

    raise IntfLvlConfErr(f"Can not find reset signal on unit {unit}")




from hwtSimApi.agents.base import AgentBase
from hwtSimApi.hdlSimulator import HdlSimulator


class StructIntfAgent(AgentBase):
    """
    Agent for StructIntf inteface

    :summary: only purpose is to instantiate agents for child interfaces
    """

    def __init__(self, sim: HdlSimulator, intf):
        AgentBase.__init__(self, sim, intf)
        for intf in intf._interfaces:
            intf._initSimAgent(sim)

    def set_data(self, d):
        intf = self.intf
        if d is None:
            for i in intf._interfaces:
                i._ag.set_data(None)
        else:
            assert len(d) == len(intf._interfaces), (d, intf._interfaces)
            for v, i in zip(d, intf._interfaces):
                i._ag.set_data(v)
    
    def get_data(self):
        intf = self.intf
        return tuple(i._ag.get_data() for i in intf._interfaces)

    def getMonitors(self):
        for intf in self.intf._interfaces:
            yield from intf._ag.getMonitors()

    def getDrivers(self):
        for intf in self.intf._interfaces:
            yield from intf._ag.getDrivers()



from typing import Union

from hwt.doc_markers import internal
from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import INT, BOOL
from hwt.hdl.types.hdlType import default_auto_cast_fn, HdlType
from hwt.hdl.types.struct import HStruct
from hwt.hdl.types.union import HUnion
from hwt.hdl.value import HValue
from hwt.interfaces.structIntf import HdlType_to_Interface
from hwt.synthesizer.exceptions import TypeConversionErr
from hwt.synthesizer.hObjList import HObjList
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal
from hwt.synthesizer.vectorUtils import iterBits, fitTo_t
from pyMathBitPrecise.bit_utils import set_bit_range, mask


@internal
def convertBits__val(self: Bits, val: "BitVal", toType: HdlType):
    if toType == BOOL:
        return val != self.getValueCls().from_py(self, 0)
    elif isinstance(toType, Bits):
        if self.signed != toType.signed:
            if self.strict_sign and bool(self.signed) != bool(toType.signed):
                raise TypeConversionErr(self, toType)
            val = val._convSign__val(toType.signed)

        w_from, w_to = self.bit_length(), toType.bit_length()
        if w_from != w_to:
            if self.strict_width:
                raise TypeConversionErr(self, toType)
            if w_from > w_to:
                # cut off some bits from value
                new_m = val.vld_mask & toType.all_mask()
            else:
                # w_from < w_to, extend the value to some bit length
                extra_mask_bits = mask(w_to - w_from)
                new_m = set_bit_range(val.vld_mask, w_from, w_to - w_from, extra_mask_bits)
            val = toType.from_py(val.val, new_m)

        if val._dtype != toType:
            # sign and width checked, only name, strict_* flags can be different
            val = toType.from_py(val.val, val.vld_mask)
        return val
    elif toType == INT:
        return INT.getValueCls()(INT, val.val,
                                 int(val._is_full_valid()))

    return default_auto_cast_fn(self, val, toType)


@internal
def convertBits(self: Bits, sigOrVal, toType: HdlType):
    """
    Cast Bit subtypes, (integers, bool, ...)
    """
    if isinstance(sigOrVal, HValue):
        return convertBits__val(self, sigOrVal, toType)
    elif toType == BOOL:
        if self.bit_length() == 1:
            v = 0 if sigOrVal._dtype.negated else 1
            return sigOrVal._eq(self.getValueCls().from_py(self, v))
    elif isinstance(toType, Bits):
        if self.bit_length() == toType.bit_length():
            if self.const is toType.const:
                return sigOrVal._convSign(toType.signed)

    return default_auto_cast_fn(self, sigOrVal, toType)


@internal
def reinterpret_bits_to_hstruct__val(val: HValue, hStructT: HStruct):
    """
    Reinterpret signal of type Bits to signal of type HStruct
    """
    container = hStructT.from_py(None)
    hStructT.vld_mask = int(val._is_full_valid())
    offset = 0
    for f in hStructT.fields:
        t = f.dtype
        width = t.bit_length()
        if f.name is not None:
            v = val[(width + offset):offset]
            v = v._reinterpret_cast(t)
            setattr(container, f.name, v)

        offset += width

    return container

@internal
def transfer_signals(src: InterfaceBase, dst: InterfaceBase):
    if src._interfaces:
        assert len(src._interfaces) == len(dst._interfaces), (src, dst)
        for si, di in zip(src._interfaces, dst._interfaces):
            transfer_signals(si, di)
    else:
        dst._sig = src._sig
        dst._sigInside = src._sigInside


@internal
def reinterpret_bits_to_hstruct(val: Union[RtlSignal, HValue], hStructT: HStruct):
    """
    Reinterpret signal of type Bits to signal of type HStruct
    """
    container = HdlType_to_Interface().apply(hStructT)
    container._loadDeclarations()
    offset = 0
    for f in hStructT.fields:
        t = f.dtype
        width = t.bit_length()
        if f.name is not None:
            v = val[(width + offset):offset]
            v = v._reinterpret_cast(t)
            current = getattr(container, f.name)
            if isinstance(v, InterfaceBase):
                transfer_signals(v, current)
            elif isinstance(v, HObjList):
                raise NotImplementedError()
            elif isinstance(v, (RtlSignal, HValue)):
                current._sig = v
            else:
                raise NotImplementedError()

        offset += width

    return container


@internal
def reinterpret_bits_to_harray(sigOrVal: Union[RtlSignal, HValue], hArrayT: HArray):
    elmT = hArrayT.element_t
    elmWidth = elmT.bit_length()
    if isinstance(sigOrVal, HValue):
        a = hArrayT.from_py(None)
        a.vld_mask = int(sigOrVal._is_full_valid())
    else:
        a = HObjList([None for _ in range(hArrayT.size)])

    for i, item in enumerate(iterBits(sigOrVal,
                                      bitsInOne=elmWidth,
                                      skipPadding=False)):
        item = item._reinterpret_cast(elmT)
        a[i] = item

    return a


@internal
def reinterpretBits__val(self: Bits, val: HValue, toType: HdlType):
    if isinstance(toType, Bits):
        if self.signed != toType.signed:
            val = val._convSign__val(toType.signed)
        return fitTo_t(val, toType)
    elif isinstance(toType, HStruct):
        return reinterpret_bits_to_hstruct__val(val, toType)
    elif isinstance(toType, HUnion):
        raise NotImplementedError()
    elif isinstance(toType, HArray):
        return reinterpret_bits_to_harray(val, toType)

    return default_auto_cast_fn(self, val, toType)


@internal
def reinterpretBits(self: Bits, sigOrVal: Union[RtlSignal, HValue], toType: HdlType):
    """
    Cast object of same bit size between to other type
    (f.e. bits to struct, union or array)
    """
    if isinstance(sigOrVal, HValue):
        return reinterpretBits__val(self, sigOrVal, toType)
    elif isinstance(toType, Bits):
        if self.signed != toType.signed:
            sigOrVal = sigOrVal._convSign(toType.signed)
        return fitTo_t(sigOrVal, toType)
    elif sigOrVal._dtype.bit_length() == toType.bit_length():
        if isinstance(toType, HStruct):
            return reinterpret_bits_to_hstruct(sigOrVal, toType)
        elif isinstance(toType, HUnion):
            raise NotImplementedError()
        elif isinstance(toType, HArray):
            return reinterpret_bits_to_harray(sigOrVal, toType)

    return default_auto_cast_fn(self, sigOrVal, toType)



"""
SystemC serializer serializes HDL objects to systemC code.
"""
from hdlConvertorAst.to.systemc._main import ToSystemc
from hwt.serializer.serializer_config import DummySerializerConfig
from hwt.serializer.systemC.serializer import ToHdlAstSystemC
from hwt.serializer.xdc.serializer import XdcSerializer


class SystemCSerializer(DummySerializerConfig):
    fileExtension = '.cpp'
    TO_HDL_AST = ToHdlAstSystemC
    TO_HDL = ToSystemc
    TO_CONSTRAINTS = XdcSerializer



from typing import Generator, Union, Tuple

from hwt.doc_markers import internal
from hwt.hdl.hdlObject import HdlObject
from hwt.hdl.operatorDefs import isEventDependentOp, OpDefinition
from hwt.hdl.sensitivityCtx import SensitivityCtx
from hwt.hdl.value import HValue
from hwt.pyUtils.arrayQuery import arr_all
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal, RtlSignalBase


@internal
def getCtxFromOps(operands):
    for o in operands:
        if isinstance(o, RtlSignalBase):
            return o.ctx
    raise TypeError("Can not find context because there is no signal in ops"
                    "(value operators should be already resolved)")


def isConst(item):
    """
    :return: True if expression is constant
    """
    return isinstance(item, HValue) or item._const


class Operator(HdlObject):
    """
    Class of operator in expression tree

    :ivar ~.operands: list of operands
    :ivar ~.evalFn: function to evaluate this operator
    :ivar ~.operator: OpDefinition instance
    :ivar ~.result: result signal of this operator
    """

    def __init__(self, operator: OpDefinition,
                 operands: Tuple[Union[RtlSignalBase, HValue]]):
        self.operands = tuple(operands)
        self.operator = operator
        self.result = None  # type: RtlSignal

    @internal
    def staticEval(self):
        """
        Recursively statistically evaluate result of this operator
        """
        for o in self.operands:
            o.staticEval()
        self.result._val = self.operator.eval(self, simulator=None)

    @internal
    def _walk_sensitivity(self, casualSensitivity: set, seen: set, ctx: SensitivityCtx):
        seen.add(self)

        if isEventDependentOp(self.operator):
            if ctx.contains_ev_dependency:
                assert self in ctx, "has to have only a single clock signal"
            ctx.contains_ev_dependency = True
            ctx.append(self)
        else:
            # walk source of signal
            for operand in self.operands:
                if operand not in seen:
                    operand._walk_sensitivity(casualSensitivity, seen, ctx)

    @internal
    def _walk_public_drivers(self, seen: set) -> Generator["RtlSignal", None, None]:
        """
        Walk all non hiden signals in an expression
        """
        for op in self.operands:
            if not isinstance(op, HValue) and op not in seen:
                seen.add(op)
                yield from op._walk_public_drivers(seen)

    @internal
    def __eq__(self, other):
        return self is other or (
            type(self) is type(other) and
            self.operator == other.operator and
            self.operands == other.operands
        )

    @internal
    @staticmethod
    def withRes(opDef, operands, resT):
        """
        Create operator with result signal

        :ivar ~.resT: data type of result signal
        :ivar ~.outputs: iterable of signals which are outputs
            from this operator
        """
        # try return existing operator result
        for i, o in enumerate(operands):
            if isinstance(o, RtlSignalBase):
                if i == 0:
                    k = (opDef, i, *operands[1:])
                else:
                    k = (opDef, i, *operands[:i], *operands[i + 1:])
                try:
                    return o._usedOps[k]
                except KeyError:
                    pass
                break

        # instanciate new Operator
        op = Operator(opDef, operands)
        out = RtlSignal(getCtxFromOps(operands), None, resT)
        out._const = arr_all(op.operands, isConst)
        out.drivers.append(op)
        out.origin = op
        op.result = out

        # Register potential signals to drivers/endpoints
        first_signal = True
        for i, o in enumerate(op.operands):
            if isinstance(o, RtlSignalBase):
                o.endpoints.append(op)
                if first_signal:
                    # register operator in _usedOps operator cache
                    if i == 0:
                        k = (opDef, i, *operands[1:])
                    else:
                        k = (opDef, i, *operands[:i], *operands[i + 1:])
                    o._usedOps[k] = out
                    o._usedOpsAlias[k] = {k, }
                    first_signal = False
            elif isinstance(o, HValue):
                pass
            else:
                raise NotImplementedError(
                    "Operator operands can be"
                    f" only signal or values got: {o}")

        if out._const:
            # if this signal is constant precompute its value
            out.staticEval()

        return out

    @internal
    def __hash__(self):
        return hash((self.operator, self.operands))

    def _destroy(self):
        self.result.drivers.remove(self)
        operands = self.operands
        first_op_sig = True
        for i, o in enumerate(operands):
            if isinstance(o, RtlSignalBase):
                # discard because operads may bethe same signal
                o.endpoints.discard(self)
                if first_op_sig:
                    # clean all references on this operator instance from RtlSignal._usedOps operator cache
                    _k = (self.operator, i, *operands[:i], *operands[i + 1:])
                    for k in o._usedOpsAlias[_k]:
                        res = o._usedOps.pop(k)
                        assert res is self.result
                    first_op_sig = False
        self.result.origin = None
        self.result = None
        self.operands = None
        self.operator = None



from hdlConvertorAst.to.verilog.constants import SIGNAL_TYPE


class SignalTypeSwap():
    """
    An object which is used as a context manager for signalType
    inside of :class:`hwt.serializer.verilog.serializer.ToHdlAstVerilog`
    """

    def __init__(self, ctx, signalType: SIGNAL_TYPE):
        self.ctx = ctx
        self.signalType = signalType

    def __enter__(self):
        self.orig = self.ctx.signalType
        self.ctx.signalType = self.signalType

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.ctx.signalType = self.orig



from hwt.synthesizer.interface import Interface
from hwt.synthesizer.unit import Unit
from ipCorePackager.constants import INTF_DIRECTION


def connect_to_const(val, intf: Interface):
    """
    Connect constant to all output ports, used mainly during the debbug
    to dissable interface
    """
    if intf._interfaces:
        for i in intf._interfaces:
            connect_to_const(val, i)
    else:
        if intf._direction == INTF_DIRECTION.SLAVE:
            intf(val)


class EmptyUnit(Unit):
    """
    :class:`hwt.synthesizer.unit.Unit` used for prototyping all output interfaces are connected
    to _def_val and this is only think which architecture contains

    :cvar _def_val: this value is used to initialize all signals
    """
    _def_val = None

    def _impl(self):
        for i in self._interfaces:
            connect_to_const(self._def_val, i)



from typing import List

from hwt.interfaces.std import Rst, Rst_n, Clk
from hwt.pyUtils.arrayQuery import where
from hwt.serializer.ip_packager import IpPackager
from hwt.synthesizer.interface import Interface

from ipCorePackager.component import Component
from ipCorePackager.intfIpMeta import IntfIpMeta, VALUE_RESOLVE


class IP_Clk(IntfIpMeta):

    def __init__(self):
        super().__init__()
        self.name = "clock"
        self.version = "1.0"
        self.vendor = "xilinx.com"
        self.library = "signal"
        self.map = 'CLK'

    def postProcess(self, component: Component,
                    packager: IpPackager, thisIf: Clk):
        rst = where(
            component.busInterfaces,
            lambda intf: (isinstance(intf, (Rst_n, Rst))
                          and intf._getAssociatedClk() is thisIf))
        logicName = packager.getInterfaceLogicalName
        rst = list(rst)
        if rst:
            if len(rst) > 1:
                rst = [intf for intf in rst if intf._getAssociatedClk() is thisIf]
                if len(rst) > 0:
                    raise AssertionError(
                        "Multiple associated resets for this interface",
                        thisIf)
                elif not rst:
                    raise AssertionError("Multiple reset signals"
                                         " but none of them is associated",
                                         thisIf)
            rst = rst[0]
            self.addSimpleParam(logicName(thisIf), "ASSOCIATED_RESET",
                                logicName(rst), resolve=VALUE_RESOLVE.NONE)
        else:
            rst = None
        associated_intfs = where(
            component.busInterfaces,
            lambda intf: (intf is not rst
                          and intf is not self
                          and not isinstance(intf, Clk)
                          and intf._getAssociatedClk() is thisIf))
        self.addSimpleParam(logicName(thisIf), "ASSOCIATED_BUSIF", ":".join(
            map(logicName, associated_intfs)), resolve=VALUE_RESOLVE.NONE)
        self.addSimpleParam(logicName(thisIf), "FREQ_HZ", str(thisIf.DEFAULT_FREQ),
                            resolve=VALUE_RESOLVE.USER)

    def asQuartusTcl(self, buff: List[str], version: str, component: Component,
                     packager: IpPackager, thisIf: Interface):
        self.quartus_tcl_add_interface(buff, thisIf, packager)
        name = packager.getInterfacePhysicalName(thisIf)
        self.quartus_prop(buff, name, "clockRate", 0)
        self.quartus_add_interface_port(buff, name, thisIf, "clk", packager)


class IP_Rst(IntfIpMeta):

    def __init__(self):
        super().__init__()
        self.name = "reset"
        self.version = "1.0"
        self.vendor = "xilinx.com"
        self.library = "signal"
        self.map = "rst"

    def postProcess(self, component: Component, packager: IpPackager, thisIf: Rst):
        self.addSimpleParam(packager.getInterfaceLogicalName(thisIf), "POLARITY", "ACTIVE_HIGH")

    def asQuartusTcl(self, buff: List[str], version: str, component: Component,
                     packager: IpPackager, thisIf: Interface):
        self.quartus_tcl_add_interface(buff, thisIf, packager)
        name = packager.getInterfacePhysicalName(thisIf)
        # self.quartus_prop("associatedClock", clock)
        self.quartus_prop(buff, name, "synchronousEdges", "DEASSERT")
        self.quartus_add_interface_port(buff, name, thisIf, "reset", packager)
        clk = thisIf._getAssociatedClk()
        if clk is not None:
            self.quartus_prop(buff, name, "associatedClock",
                              packager.getInterfacePhysicalName(clk),
                              escapeStr=False)


class IP_Rst_n(IP_Rst):

    def postProcess(self, component: Component, packager: IpPackager, thisIf: Rst_n):
        self.addSimpleParam(
            packager.getInterfaceLogicalName(thisIf), "POLARITY", "ACTIVE_LOW")

    def asQuartusTcl(self, buff: List[str], version: str, component: Component,
                     packager: IpPackager, thisIf: Interface):
        self.quartus_tcl_add_interface(buff, thisIf, packager)
        name = packager.getInterfacePhysicalName(thisIf)
        # [TODO]
        # self.quartus_prop("associatedClock", clock)
        self.quartus_prop(buff, name, "synchronousEdges", "DEASSERT")
        self.quartus_add_interface_port(buff, name, thisIf, "reset_n", packager)
        clk = thisIf._getAssociatedClk()
        if clk is not None:
            self.quartus_prop(buff, name, "associatedClock",
                              packager.getInterfacePhysicalName(clk),
                              escapeStr=False)


class IP_Handshake(IntfIpMeta):

    def __init__(self):
        super().__init__()
        self.name = "handshake"
        self.version = "1.0"
        self.vendor = "hwt"
        self.library = "user"
        self.map = {
            "vld": "valid",
            "rd": "ready",
            "data": "data"
        }


class IP_BlockRamPort(IntfIpMeta):

    def __init__(self):
        super().__init__()
        self.name = "bram"
        self.version = "1.0"
        self.vendor = "xilinx.com"
        self.library = "interface"
        self.map = {
            'addr': "ADDR",
            "clk": 'CLK',
            'din': "DIN",
            'dout': "DOUT",
            'en': "EN",
            'we': "WE",
        }


class IP_FifoReader(IntfIpMeta):

    def __init__(self):
        super().__init__()
        self.name = "fifo_read"
        self.version = "1.0"
        self.vendor = "xilinx.com"
        self.library = "interface"
        self.map = {
            'data': "RD_DATA",
            "en": 'RD_EN',
            'wait': "EMPTY",
        }


class IP_FifoWriter(IntfIpMeta):

    def __init__(self):
        super().__init__()
        self.name = "fifo_write"
        self.version = "1.0"
        self.vendor = "xilinx.com"
        self.library = "interface"
        self.map = {
            'data': "WR_DATA",
            "en": 'WR_EN',
            'wait': "FULL",
        }


from hwt.hdl.constants import INTF_DIRECTION, DIRECTION
from hwt.doc_markers import internal


@internal
class InterfaceDirectionFns():

    @internal
    def _setDirectionsLikeIn(self, intfDir):
        assert intfDir in [INTF_DIRECTION.MASTER,
                           INTF_DIRECTION.SLAVE,
                           INTF_DIRECTION.TRISTATE], intfDir
        d = DIRECTION.asIntfDirection(self._masterDir)
        if intfDir == INTF_DIRECTION.MASTER or d == INTF_DIRECTION.TRISTATE:
            pass
        else:
            d = INTF_DIRECTION.opposite(d)

        self._direction = d
        for i in self._interfaces:
            i._setDirectionsLikeIn(d)

    @internal
    def _setAsExtern(self, isExtern):
        """Set interface as extern"""
        self._isExtern = isExtern
        for prop in self._interfaces:
            prop._setAsExtern(isExtern)

    @internal
    def _reverseDirection(self):
        """Reverse direction of this interface in implementation stage"""
        self._direction = INTF_DIRECTION.opposite(self._direction)
        for intf in self._interfaces:
            intf._reverseDirection()



from copy import copy, deepcopy
from typing import Optional, List

from hdlConvertorAst.hdlAst import HdlModuleDef, HdlStmIf, HdlModuleDec, HdlOp, \
    HdlOpType, HdlValueId
from hdlConvertorAst.hdlAst._defs import HdlIdDef
from hdlConvertorAst.hdlAst._expr import HdlTypeAuto
from hdlConvertorAst.hdlAst._statements import HdlStmProcess, HdlStmBlock, HdlStmAssign, \
    HdlStmWait
from hdlConvertorAst.to.verilog.keywords import IEEE1800_2017_KEYWORDS
from hdlConvertorAst.translate.common.name_scope import LanguageKeyword, NameScope
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_call
from hwt.hdl.portItem import HdlPortItem
from hwt.hdl.types.array import HArray
from hwt.hdl.types.defs import STR, INT, BOOL
from hwt.serializer.generic.to_hdl_ast import ToHdlAst
from hwt.serializer.verilog.context import SignalTypeSwap
from hwt.serializer.verilog.ops import ToHdlAstVerilog_ops
from hwt.serializer.verilog.statements import ToHdlAstVerilog_statements
from hwt.serializer.verilog.types import ToHdlAstVerilog_types
from hwt.serializer.verilog.utils import SIGNAL_TYPE, verilogTypeOfSig
from hwt.serializer.verilog.value import ToHdlAstVerilog_Value


class ToHdlAstVerilog(ToHdlAstVerilog_types,
                      ToHdlAstVerilog_Value, ToHdlAstVerilog_statements,
                      ToHdlAstVerilog_ops, ToHdlAst):
    _keywords_dict = {kw: LanguageKeyword() for kw in IEEE1800_2017_KEYWORDS}

    def __init__(self, name_scope: Optional[NameScope]=None):
        ToHdlAst.__init__(self, name_scope=name_scope)
        self.signalType = SIGNAL_TYPE.PORT_WIRE

    def as_hdl_HdlModuleDef_variable(
            self, v, types, hdl_types, hdl_variables,
            processes, component_insts):
        new_v = copy(v)
        with SignalTypeSwap(self, verilogTypeOfSig(v.origin)):
            t = v.type
            # if type requires extra definition
            if self.does_type_requires_extra_def(t, types):
                _t = self.as_hdl_HdlType(t, declaration=True)
                hdl_types.append(_t)
                types.add(t)
            new_v.type = self.as_hdl_HdlType(t, declaration=False)
            # this is a array variable which requires value intialization in init
            # process
            if isinstance(t, HArray):
                if v.value.vld_mask:
                    rom = v.origin
                    p = HdlStmProcess()
                    label = self.name_scope.checked_name(rom.name + "_rom_init", p)
                    p.labels.append(label)
                    p.body = HdlStmBlock()
                    body = p.body.body
                    for i, _v in enumerate(rom.def_val.val):
                        a = HdlStmAssign(self.as_hdl_int(int(_v)),
                                         self.as_hdl(rom[i]))
                        a.is_blocking = True
                        body.append(a)
                    w = HdlStmWait()
                    w.val = []  # initial process
                    body.append(w)
                    processes.append(p)
                # because we would not be able to initialize const/localparam variable later
                new_v.is_const = False
                new_v.value = None
            elif new_v.value is not None:
                if new_v.value.vld_mask:
                    new_v.value = self.as_hdl_Value(new_v.value)
                else:
                    # 'x' is a default value no need to specify it extra
                    new_v.value = None
            return new_v

    def _as_hdl_HdlModuleDef_param_asserts(self, new_m: HdlModuleDec) -> List[HdlOp]:
        res = []
        for p in new_m.params:
            p: HdlIdDef
            if p.value is None:
                continue
            # [TODO] this is actually requires SV>=2009
            # generate
            # if (p==x) begin
            #     $error("%m Generated only for this param value");
            # end
            # endgenerate
            i = HdlStmIf()
            i.in_preproc = True
            i.cond = HdlOp(HdlOpType.NE, [HdlValueId(p.name), deepcopy(p.value)])
            i.if_true = hdl_call(HdlValueId("$error"), [
                 "%m Generated only for this param value",
                 ])
            res.append(i)

        return res

    def as_hdl_HdlModuleDef(self, o: HdlModuleDef):
        o = super(ToHdlAstVerilog, self).as_hdl_HdlModuleDef(o)
        param_asserts = self._as_hdl_HdlModuleDef_param_asserts(o.dec)

        o.objs.extend(param_asserts)
        return o

    def as_hdl_GenericItem(self, g: HdlIdDef):
        with SignalTypeSwap(self, SIGNAL_TYPE.PORT_WIRE):
            new_v = copy(g)
            v = g.value
            if v._dtype == STR or v._dtype == INT or v._dtype == BOOL:
                t = HdlTypeAuto
            else:
                t = self.as_hdl_HdlType(v._dtype)
            new_v.type = t
            assert new_v.value is not None, g
            new_v.value = self.as_hdl_Value(v)
            return new_v

    def as_hdl_HdlPortItem(self, pi: HdlPortItem):
        with SignalTypeSwap(self, verilogTypeOfSig(pi)):
            v = super(ToHdlAstVerilog, self).as_hdl_HdlPortItem(pi)
            v.is_latched = self.signalType == SIGNAL_TYPE.PORT_REG
        return v




from io import StringIO
from hdlConvertorAst.translate.common.name_scope import NameScope


class HdlObject():
    """
    Base Hdl object class for object which can be directly serialized
    to target HDL language
    """

    def __repr__(self):
        from hwt.serializer.hwt import HwtSerializer
        name_scope = NameScope(None, "debug", False, debug=True)
        to_hdl = HwtSerializer.TO_HDL_AST(name_scope)
        to_hdl.debug = True
        hdl = to_hdl.as_hdl(self)
        buff = StringIO()
        # import sys
        # buff = sys.stdout
        ser = HwtSerializer.TO_HDL(buff)
        ser.visit_iHdlObj(hdl)
        return buff.getvalue()



from typing import Tuple, Callable, Union

from pyMathBitPrecise.array3t import Array3t
from pyMathBitPrecise.bits3t import Bits3t
from pyMathBitPrecise.enum3t import Enum3t

from hwt.doc_markers import internal
from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.enum import HEnum
from hwt.hdl.types.hdlType import HdlType
from hwt.hdl.value import HValue
from hwt.simulator.rtlSimulator import BasicRtlSimulatorWithSignalRegisterMethods
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from pyDigitalWaveTools.json.writer import JsonWriter
from pyDigitalWaveTools.vcd.common import VCD_SIG_TYPE
from hwtSimApi.basic_hdl_simulator.proxy import BasicRtlSimProxy
from hwtSimApi.basic_hdl_simulator.sim_utils import ValueUpdater, \
    ArrayValueUpdater
from pyDigitalWaveTools.json.value_format import JsonBitsFormatter,\
    JsonEnumFormatter, JsonArrayFormatter


class BasicRtlSimulatorJson(BasicRtlSimulatorWithSignalRegisterMethods):
    supported_type_classes = (Bits, HEnum, HArray, Bits3t, Enum3t, Array3t)

    @internal
    def get_trace_formatter(self, t: HdlType)\
            -> Tuple[str, int, Callable[[RtlSignalBase, HValue], str]]:
        """
        :return: (vcd type name, vcd width, formatter fn)
        """
        if isinstance(t, (Bits3t, Bits)):
            return (VCD_SIG_TYPE.WIRE, t.bit_length(), JsonBitsFormatter())
        elif isinstance(t, (Enum3t, HEnum)):
            return (VCD_SIG_TYPE.ENUM, 1, JsonEnumFormatter())
        elif isinstance(t, (HArray, Array3t)):
            dimensions = []
            while isinstance(t, (HArray, Array3t)):
                dimensions.append(t.size)
                t = t.element_t
            _, _, elm_format = self.get_trace_formatter(t)
            return (VCD_SIG_TYPE.ARRAY,
                    dimensions + [t.bit_length(), ],
                    JsonArrayFormatter(dimensions, elm_format))
        else:
            raise ValueError(t)

    def create_wave_writer(self, data):
        self.wave_writer = JsonWriter(data)

    def logChange(self, nowTime: int,
                  sig: BasicRtlSimProxy, 
                  nextVal: HValue,
                  valueUpdater: Union[ValueUpdater, ArrayValueUpdater]):
        """
        This method is called for every value change of any signal.
        """
        try:
            self.wave_writer.logChange(nowTime, sig, nextVal, valueUpdater)
        except KeyError:
            # not every signal has to be registered
            # (if it is not registered it means it is ignored)
            pass



class ValueWidthRequirementScope():
    """
    Context manager which temporarily swaps the _valueWidthRequired on specified context 

    .. code-block:: python

        with ValueWidthRequirementScope(ctx, True):
            #...
    """

    def __init__(self, ctx, val):
        self.ctx = ctx
        self.val = val

    def __enter__(self):
        self.orig = self.ctx._valueWidthRequired
        self.ctx._valueWidthRequired = self.val

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.ctx._valueWidthRequired = self.orig



from collections import deque

from hwt.simulator.agentBase import SyncAgentBase
from hwt.synthesizer.exceptions import IntfLvlConfErr
from hwtSimApi.agents.base import AgentBase
from hwtSimApi.hdlSimulator import HdlSimulator
from hwtSimApi.triggers import Timer, WaitWriteOnly, WaitCombRead, WaitCombStable
from hwtSimApi.constants import CLK_PERIOD


class SignalAgent(SyncAgentBase):
    """
    Agent for signal interface, it can use clock and reset interface
    for synchronization or can be synchronized by delay

    :attention: clock synchronization has higher priority
    """

    def __init__(self, sim: HdlSimulator, intf: "Signal", delay=None):
        AgentBase.__init__(self, sim, intf)
        self.delay = delay
        self.initDelay = 0

        # resolve clk and rstn
        try:
            self.clk = self.intf._getAssociatedClk()
        except IntfLvlConfErr:
            self.clk = None
        self.rst, self.rstOffIn = self._discoverReset(intf, True)
        self.data = deque()

        self.initPending = True

        if self.clk is None:
            if self.delay is None:
                self.delay = CLK_PERIOD
            self.monitor = self.monitorWithTimer
            self.driver = self.driverWithTimer
        else:
            if self.initDelay:
                raise NotImplementedError("initDelay only without clock")
            if self.delay:
                raise ValueError("clock and delay synchronization at once")
            c = self.SELECTED_EDGE_CALLBACK
            self.monitor = c(sim, self.clk, self.monitorWithClk, self.getEnable)
            self.driver = c(sim, self.clk, self.driverWithClk, self.getEnable)

    def getDrivers(self):
        d = SyncAgentBase.getDrivers(self)
        return [self.driverInit()] + d

    def driverInit(self):
        yield WaitWriteOnly()
        if not self._enabled:
            return
        try:
            d = self.data[0]
        except IndexError:
            d = None

        self.set_data(d)

    def get_data(self):
        return self.intf.read()

    def set_data(self, data):
        self.intf.write(data)

    def driverWithClk(self):
        # if clock is specified this function is periodically called every
        # clk tick, if agent is enabled
        yield WaitCombRead()
        if not self._enabled:
            return
        if self.data and self.notReset():
            yield WaitWriteOnly()
            if not self._enabled:
                return
            d = self.data.popleft()
            self.set_data(d)

    def driverWithTimer(self):
        if self.initPending:
            if self.initDelay:
                yield Timer(self.initDelay)
            self.initPending = False
        # if clock is specified this function is periodically called every
        # clk tick
        while True:
            yield WaitWriteOnly()
            if self._enabled and self.data and self.notReset():
                yield WaitWriteOnly()
                if self._enabled:
                    d = self.data.popleft()
                    self.set_data(d)

            yield Timer(self.delay)

    def monitorWithTimer(self):
        if self.initPending and self.initDelay:
            yield Timer(self.initDelay)
            self.initPending = False
        # if there is no clk, we have to manage periodic call by our self
        while True:
            yield WaitCombRead()
            if self._enabled and self.notReset():
                d = self.get_data()
                self.data.append(d)

            yield Timer(self.delay)

    def monitorWithClk(self):
        # if clock is specified this function is periodically called every
        # clk tick, when agent is enabled
        yield WaitCombStable()
        if self._enabled and self.notReset():
            d = self.get_data()
            self.data.append(d)



from typing import Optional, Union, Tuple

from hdlConvertorAst.hdlAst import HdlIdDef
from hdlConvertorAst.translate.common.name_scope import NameScope
from hwt.hdl.statements.assignmentContainer import HdlAssignmentContainer
from hwt.hdl.types.hdlType import HdlType
from hwt.hdl.value import HValue
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal


class TmpVarConstructor():

    def __init__(self, toHdlAst, name_scope: NameScope):
        self.toHdlAst = toHdlAst
        self.name_scope = name_scope
        self.extraVarsHdl = []
        self.cache = {}

    def create_var_cached(self, suggestedName: str, dtype: HdlType,
                       const=False,
                       def_val: Optional[Union[RtlSignalBase, HValue]]=None,
                       postponed_init=False,
                       extra_args=None) -> Tuple[bool, RtlSignal]:
        cache_k = (suggestedName, dtype, const, def_val, extra_args)
        try:
            return False, self.cache[cache_k]
        except KeyError:
            pass
        v = self.create_var(suggestedName, dtype, const=const, def_val=def_val, postponed_init=postponed_init)
        self.cache[cache_k] = v
        return True, v

    def create_var(self,
               suggestedName: str,
               dtype: HdlType,
               const=False,
               def_val: Optional[Union[RtlSignalBase, HValue]]=None,
               postponed_init=False) -> RtlSignal:
        # create a new tmp variable in current process
        s = RtlSignal(None, None, dtype, virtual_only=True)
        s.name = self.name_scope.checked_name(suggestedName, s)
        s.hidden = False
        s._const = const

        if def_val is not None:
            s.def_val = def_val
            if not (isinstance(def_val, RtlSignalBase) and not def_val._const):
                s._set_def_val()

        if not postponed_init:
            self.finish_var_init(s)

        return s

    def finish_var_init(self, var: RtlSignal):
        hdl = self.extraVarsHdl

        if isinstance(var.def_val, RtlSignalBase) or var.def_val.vld_mask:
            a = HdlAssignmentContainer(var.def_val, var, virtual_only=True)
            hdl.append(self.toHdlAst.as_hdl_HdlAssignmentContainer(a))
        else:
            assert var._const or var.drivers, (var, var.def_val)

        as_hdl = self.toHdlAst.as_hdl_SignalItem(var, declaration=True)

        for d in var.drivers:
            hdl.append(self.toHdlAst.as_hdl(d))

        hdl.append(as_hdl)

    def sort_hdl_declarations_first(self):
        self.extraVarsHdl.sort(key=lambda x: not isinstance(x, HdlIdDef))


class NoTmpVars():

    def create_var_cached(self, suggestedName, dtype, *args, **kwargs):
        raise NotImplementedError(
            "Can not create a tmp variable (%s of type %r) in this code section" % (suggestedName, dtype))

    def create_cached(self, suggestedName, dtype, *args, **kwargs):
        raise NotImplementedError(
            "Can not create a tmp variable (%s of type %r) in this code section" % (suggestedName, dtype))




from itertools import islice
from typing import Union, Tuple

from hwt.constraints import set_max_delay, set_false_path, \
    set_async_reg, get_clock_of, iHdlConstrain
from hwt.hdl.types.bits import Bits
from hwt.pyUtils.arrayQuery import iter_with_last
from hwt.synthesizer.interface import Interface
from hwt.synthesizer.rtlLevel.rtlSyncSignal import RtlSyncSignal
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal
from hwt.synthesizer.unit import Unit


class XdcSerializer():
    """
    Convert constrains containers to a XDC format (For Xilinx Vivado)
    """
    fileExtension = ".xdc"

    def __init__(self, out):
        self.out = out

    def _get(self, o: Union[Tuple[Unit, RtlSignal, Interface], iHdlConstrain], only_first=False):
        """
        :param only_first: if true select only first bit from vector, else select whole vector
        """
        if isinstance(o, iHdlConstrain):
            return self.visit_iHdlConstrain(o)

        is_reg = False
        _o = o[-1]
        if isinstance(_o, RtlSignal):
            q = "get_cells"
            for d in _o.drivers:
                if d._event_dependent_from_branch is not None:
                    is_reg = True

        elif isinstance(_o, Interface):
            q = "get_pins"

        else:
            raise NotImplementedError(o)

        w = self.out.write
        w(q)
        w(" -hier -filter {NAME =~ */")
        path = o
        # [TODO] find out how to make select with ip top module/entity name
        for last, p in iter_with_last(islice(path, 1, None)):
            if isinstance(p, Unit):
                w(p._name)
                w("_inst")
            elif isinstance(p, RtlSignal):
                w(p.name)
            elif isinstance(p, Interface):
                w(p._name)
            else:
                raise NotImplementedError(p)
            if not last:
                w("/")

        t = _o._dtype
        if is_reg:
            w("_reg")
        if isinstance(t, Bits) and (t.bit_length() > 1 or t.force_vector):
            # * on end because of Vivado _replica
            if only_first:
                w("[0]*")
            else:
                w("[*]*")
        w("}")

    def visit_get_clock_of(self, o: get_clock_of):
        w = self.out.write
        if isinstance(o.obj[-1], RtlSyncSignal):
            w("get_clocks -of [")
            self._get(o.obj)
            w("]")
        else:
            raise NotImplementedError()

    def visit_iHdlConstrain(self, o):
        visitFn = getattr(self, "visit_" + o.__class__.__name__, None)
        if visitFn is None:
            return o.to_xdc(self, o)
        else:
            return visitFn(o)

    def visit_HdlConstraintList(self, o_list):
        return [self.visit_iHdlConstrain(o) for o in o_list]

    def visit_set_async_reg(self, o: set_async_reg):
        w = self.out.write
        w("set_property ASYNC_REG TRUE [")
        self._get(o.sig)
        w("]\n")

    def visit_set_false_path(self, o: set_false_path):
        w = self.out.write
        w("set_false_path")
        if o.start is not None:
            w(" -from [")
            self._get(o.start)
            w("]")
        if o.end is not None:
            w(" -to [")
            self._get(o.end)
            w("]")
        w("\n")

    def visit_set_max_delay(self, o: set_max_delay):
        w = self.out.write
        w("set_max_delay -from [")
        self._get(o.start)
        w("] -to [")
        self._get(o.end)
        w("]")
        if o.datapath_only:
            w(" -datapath_only")
        w(f" {o.time_ns:f}\n")



from typing import Union

from hdlConvertorAst.hdlAst import iHdlStatement
from hdlConvertorAst.to.verilog.constants import SIGNAL_TYPE
from hwt.doc_markers import internal
from hwt.hdl.portItem import HdlPortItem
from hwt.hdl.statements.assignmentContainer import HdlAssignmentContainer
from hwt.hdl.value import HValue
from hwt.hdl.variables import SignalItem
from ipCorePackager.constants import DIRECTION


@internal
def verilogTypeOfSig(s: Union[SignalItem, HdlPortItem]):
    """
    Check if is register or wire
    """
    if isinstance(s, HdlPortItem):
        if s.direction == DIRECTION.IN or s.direction == DIRECTION.INOUT:
            return SIGNAL_TYPE.PORT_WIRE

        t = verilogTypeOfSig(s.getInternSig())
        if t == SIGNAL_TYPE.WIRE:
            return SIGNAL_TYPE.PORT_WIRE
        elif t == SIGNAL_TYPE.REG:
            return SIGNAL_TYPE.PORT_REG
        else:
            raise ValueError(t)

    driver_cnt = len(s.drivers)
    if driver_cnt == 1:
        d = s.drivers[0]
        if isinstance(d, HdlPortItem):
            # input port
            return SIGNAL_TYPE.WIRE
        elif isinstance(d, HdlAssignmentContainer)\
                and d.parentStm is None\
                and not d.indexes\
                and d._event_dependent_from_branch is None\
                and (isinstance(d.src, HValue) or not d.src.hidden):
            # primitive assignment
            return SIGNAL_TYPE.WIRE
        elif isinstance(d, iHdlStatement) and d.in_preproc:
            return SIGNAL_TYPE.WIRE

    return SIGNAL_TYPE.REG



from copy import copy
from typing import Optional

from hdlConvertorAst.hdlAst import iHdlStatement, iHdlObj, HdlIdDef, \
    HdlValueId, HdlTypeType, iHdlExpr, HdlStmBlock, HdlStmIf, HdlStmCase, \
    HdlStmProcess, HdlStmAssign, HdlModuleDef, HdlModuleDec, \
    HdlCompInst, HdlEnumDef
from hdlConvertorAst.hdlAst._statements import ALL_STATEMENT_CLASSES
from hdlConvertorAst.to.basic_hdl_sim_model._main import ToBasicHdlSimModel
from hdlConvertorAst.translate.common.name_scope import NameScope, WithNameScope
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import \
    hdl_index, hdl_map_asoc
from hwt.hdl.operator import Operator
from hwt.hdl.portItem import HdlPortItem
from hwt.hdl.statements.assignmentContainer import HdlAssignmentContainer
from hwt.hdl.statements.codeBlockContainer import HdlStmCodeBlockContainer
from hwt.hdl.statements.ifContainter import IfContainer
from hwt.hdl.statements.switchContainer import SwitchContainer
from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import STR, BOOL
from hwt.hdl.types.enum import HEnum
from hwt.hdl.types.float import HFloat
from hwt.hdl.types.hdlType import HdlType, MethodNotOverloaded
from hwt.hdl.types.slice import HSlice
from hwt.pyUtils.arrayQuery import arr_any
from hwt.serializer.exceptions import SerializerException
from hwt.serializer.exceptions import UnsupportedEventOpErr
from hwt.serializer.generic.tmpVarConstructor import TmpVarConstructor, \
    NoTmpVars
from hwt.serializer.generic.utils import HWT_TO_HDLCONVEROTR_DIRECTION, \
    TmpVarsSwap
from hwt.serializer.utils import HdlStatement_sort_key
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class ToHdlAst():
    """
    Base class for translators which translates hwt AST to a HDL ast

    :ivar ~.name_scope: name scope for resolution of hdl names for objects
        and for name colision checking for newly generated objects
    :ivar ~.tmpVars: A object which is used to create a tmp variable in current scope.
        It is set if it is possible to create a tmp variable.
    :ivar ~.constCache: A ConstCache instance used o extract values as a constants.
    :type ~.constCache: Optional[ConstantCache]
    """
    # used to filter statems from other object by class without using
    # isisnstance
    ALL_STATEMENT_CLASSES = [*ALL_STATEMENT_CLASSES, HdlStmCodeBlockContainer]
    TMP_VAR_CONSTRUCTOR = TmpVarConstructor
    _keywords_dict = {}

    @classmethod
    def getBaseNameScope(cls):
        """
        Get root of name space
        """
        s = NameScope.make_top(False)
        s.update(cls._keywords_dict)
        return s

    def __init__(self, name_scope: Optional[NameScope]=None):
        if name_scope is None:
            name_scope = self.getBaseNameScope()
        self.name_scope = name_scope
        self.tmpVars = NoTmpVars()
        self.constCache = None

    def as_hdl(self, obj) -> iHdlObj:
        """
        Convert any object to HDL AST

        :param obj: object to convert
        """
        serFn = getattr(self, "as_hdl_" + obj.__class__.__name__, None)
        if serFn is not None:
            return serFn(obj)
        elif isinstance(obj, RtlSignalBase):
            return self.as_hdl_SignalItem(obj)
        else:
            raise SerializerException(self,
                                      "Not implemented for obj of",
                                      obj.__class__, obj)

    def as_hdl_HdlType(self, typ: HdlType, declaration=False):
        try:
            return typ._as_hdl(self, declaration)
        except MethodNotOverloaded:
            pass

        if typ == STR:
            sFn = self.as_hdl_HdlType_str
        elif isinstance(typ, Bits):
            sFn = self.as_hdl_HdlType_bits
        elif isinstance(typ, HEnum):
            sFn = self.as_hdl_HdlType_enum
        elif isinstance(typ, HArray):
            sFn = self.as_hdl_HdlType_array
        elif isinstance(typ, HSlice):
            sFn = self.as_hdl_HdlType_slice
        elif isinstance(typ, HFloat):
            sFn = self.as_hdl_HdlType_float
        else:
            # [todo] better error msg
            raise NotImplementedError("type declaration is not implemented"
                                      " for type %s"
                                      % (HdlType.__repr__(typ)))

        return sFn(typ, declaration=declaration)

    def as_hdl_HdlType_array(self, typ: HArray, declaration=False):
        ns = self.name_scope
        if declaration:
            dec = HdlIdDef()
            dec.type = HdlTypeType
            if self.does_type_requires_extra_def(typ.element_t, ()):
                # problem there is that we do not have a list of already defined types
                # so we can not just declare an element type
                raise NotImplementedError(typ.element_t)

            dec.value = hdl_index(self.as_hdl_HdlType(typ.element_t, declaration=False),
                                  self.as_hdl_int(int(typ.size)))
            name = getattr(typ, "name", "arr_t_")
            dec.name = ns.checked_name(name, typ)
            return dec
        else:
            name = ns.get_object_name(typ)
            return HdlValueId(name, obj=typ)

    def as_hdl_HdlType_enum(self, typ: HEnum, declaration=False):
        ns = self.name_scope
        if declaration:
            e = HdlEnumDef()
            e.origin = typ
            e.name = ns.checked_name(typ.name, typ)
            e.values = [(ns.checked_name(n, getattr(typ, n)), None)
                        for n in typ._allValues]
            dec = HdlIdDef()
            dec.type = HdlTypeType
            dec.value = e
            dec.name = e.name
            return dec
        else:
            name = ns.get_object_name(typ)
            return HdlValueId(name, obj=None)

    def as_hdl_HdlType_slice(self, typ: HSlice, declaration=False):
        raise NotImplementedError(self)

    def as_hdl_If(self, *args, **kwargs) -> HdlStmIf:
        return self.as_hdl_IfContainer(*args, **kwargs)

    def as_hdl_cond(self, v, force_bool) -> iHdlExpr:
        if force_bool and v._dtype != BOOL:
            v = v._isOn()
        return self.as_hdl(v)

    def as_hdl_statements(self, stm_list) -> iHdlStatement:
        if stm_list is None:
            return None
        elif len(stm_list) == 1:
            return self.as_hdl(stm_list[0])
        else:
            b = HdlStmBlock()
            b.body = [self.as_hdl(s) for s in stm_list]
            return b

    def _as_hdl_HdlAssignmentContainer_auto_conversions(self, a: HdlAssignmentContainer):
        dst = a.dst
        src = a.src
        dst_indexes = a.indexes
        if a.indexes is not None:
            dst_indexes = [self.as_hdl(x) for x in dst_indexes]
            correct = True
        else:
            if dst._dtype == a.src._dtype:
                correct = True
            else:
                srcT = a.src._dtype
                dstT = dst._dtype
                correct = False
                if (isinstance(srcT, Bits) and
                        isinstance(dstT, Bits)):
                    bl0 = srcT.bit_length()
                    if bl0 == dstT.bit_length():
                        if bl0 == 1 and srcT.force_vector != dstT.force_vector:
                            if srcT.force_vector:
                                src = src[0]
                                correct = True
                            elif dstT.force_vector:
                                dst_indexes = [self.as_hdl_int(0), ]
                                correct = True
                        elif srcT.signed == dstT.signed:
                            correct = True

        if not correct:
            raise SerializerException((
                "%s <= %s  is not valid assignment\n"
                " because types are different (%r; %r) ") %
                (dst, src, dst._dtype, a.src._dtype))
        return dst, dst_indexes, self.as_hdl_Value(src)

    def as_hdl_HdlAssignmentContainer(self, a: HdlAssignmentContainer):
        dst, dst_indexes, src = self._as_hdl_HdlAssignmentContainer_auto_conversions(a)
        dst = self.as_hdl(dst)
        if dst_indexes:
            for i in dst_indexes:
                dst = hdl_index(dst, i)
        a = HdlStmAssign(src, dst)
        return a

    def as_hdl_IfContainer(self, ifc: IfContainer) -> HdlStmIf:
        try:
            cond = self.as_hdl_cond(ifc.cond, True)
        except UnsupportedEventOpErr:
            cond = None

        if cond is None:
            assert not ifc.elIfs
            assert not ifc.ifFalse
            return self.as_hdl_statements(ifc.ifTrue)

        elIfs = []
        ifTrue = self.as_hdl_statements(ifc.ifTrue)
        ifFalse = self.as_hdl_statements(ifc.ifFalse)

        for c, statements in ifc.elIfs:
            try:
                elIfs.append((self.as_hdl_cond(c, True),
                              self.as_hdl_statements(statements)))
            except UnsupportedEventOpErr:
                if len(ifc.elIfs) == 1 and not ifFalse:
                    # register expression is in valid format and this
                    # is just register with asynchronous reset or etc...
                    ifFalse = self.as_hdl_statements(statements)
                else:
                    raise

        i = HdlStmIf()
        i.cond = cond
        i.if_true = ifTrue
        i.elifs = elIfs
        i.if_false = ifFalse
        return i

    def as_hdl_Switch(self, *args, **kwargs) -> HdlStmCase:
        return self.as_hdl_SwitchContainer(*args, **kwargs)

    def as_hdl_FsmBuilder(self, *args, **kwargs) -> HdlStmCase:
        return self.as_hdl_SwitchContainer(*args, **kwargs)

    def as_hdl_SwitchContainer(self, sw: SwitchContainer) -> HdlStmCase:
        s = HdlStmCase()
        s.switch_on = self.as_hdl_cond(sw.switchOn, False)
        s.cases = cases = []
        for key, statements in sw.cases:
            key = self.as_hdl_Value(key)
            cases.append((key, self.as_hdl_statements(statements)))

        s.default = self.as_hdl_statements(sw.default)
        return s

    def as_hdl_PortConnection(self, o: HdlPortItem):
        assert isinstance(o, HdlPortItem), o
        if o.dst._dtype != o.src._dtype:
            raise SerializerException(
                f"Port map {o.name:s} is not valid (types does not match)  ({o.src._dtype}, {o.dst._dtype}) "
                f"{o.src} => {o.dst}"
            )

        intern, outer = o.getInternSig(), o.getOuterSig()
        intern_hdl = self.as_hdl_Value(intern)
        intern_hdl.obj = o
        outer_hdl = self.as_hdl_Value(outer)
        pm = hdl_map_asoc(intern_hdl, outer_hdl)
        return pm

    def as_hdl_HdlCompInst(self, o: HdlCompInst) -> HdlCompInst:
        new_o = copy(o)
        param_map = []
        for p in o.param_map:
            assert isinstance(p, HdlIdDef), p
            pm = hdl_map_asoc(HdlValueId(p.name, obj=p), self.as_hdl(p.value))
            param_map.append(pm)
        new_o.param_map = param_map

        port_map = []
        for pi in o.port_map:
            pm = self.as_hdl_PortConnection(pi)
            port_map.append(pm)
        new_o.port_map = port_map
        return new_o

    def as_hdl_GenericItem(self, o: HdlIdDef):
        assert not self.does_type_requires_extra_def(o.type, tuple())
        return self.as_hdl_HdlModuleDef_variable(o, None, None, None, None, None)

    def as_hdl_HdlPortItem(self, o: HdlPortItem):
        var = HdlIdDef()
        var.direction = HWT_TO_HDLCONVEROTR_DIRECTION[o.direction]
        s = o.getInternSig()
        var.name = s.name
        var.origin = o
        var.type = o._dtype
        return self.as_hdl_HdlModuleDef_variable(var, (), None, None, None, None)

    def as_hdl_HdlModuleDec(self, o: HdlModuleDec):
        # :attention: name_scope should be already set to body of module
        # with WithNameScope(self, self.name_scope.get_child(o.name)):

        new_o = copy(o)

        # convert types, exprs
        new_o.params = [self.as_hdl_GenericItem(p) for p in o.params]
        new_o.ports = [self.as_hdl_HdlPortItem(p) for p in o.ports]

        return new_o

    def does_type_requires_extra_def(self, t: HdlType, other_types: list):
        try:
            return t._as_hdl_requires_def(self, other_types)
        except MethodNotOverloaded:
            pass
        return isinstance(t, (HEnum, HArray)) and t not in other_types

    def as_hdl_HdlModuleDef_variable(
            self, v, types, hdl_types, hdl_variables,
            processes, component_insts):
        t = v.type
        # if type requires extra definition
        if self.does_type_requires_extra_def(t, types):
            _t = self.as_hdl_HdlType(t, declaration=True)
            hdl_types.append(_t)
            types.add(t)

        return self.as_hdl_SignalItem(v, declaration=True)

    def _as_hdl_HdlModuleDef(self, new_m: HdlModuleDef) -> HdlModuleDef:
        # with WithNameScope(self,
        # self.name_scope.get_child(o.module_name.val)):
        hdl_types, hdl_variables, processes, component_insts = \
            ToBasicHdlSimModel.split_HdlModuleDefObjs(self, new_m.objs)
        # [TODO] sorting not required as it should be done in _to_rtl()
        if len(hdl_variables) > 1:
            hdl_variables.sort(key=lambda x: (x.name, x.origin._instId))
        if len(processes) > 1:
            processes.sort(key=HdlStatement_sort_key)
        if len(component_insts) > 1:
            component_insts.sort(key=lambda x: x.name)

        types = set()

        extraVars = self.TMP_VAR_CONSTRUCTOR(self, self.name_scope)

        with TmpVarsSwap(self, extraVars):
            return self._as_hdl_HdlModuleDef_body(
                new_m, types, hdl_types, hdl_variables, extraVars,
                processes, component_insts)

    def _as_hdl_HdlModuleDef_body(
            self, new_m, types, hdl_types, hdl_variables,
            extraVars: TmpVarConstructor, processes, component_insts):

        _hdl_variables = []
        for v in hdl_variables:
            new_v = self.as_hdl_HdlModuleDef_variable(
                v, types, hdl_types, hdl_variables,
                processes, component_insts)
            _hdl_variables.append(new_v)
        hdl_variables = _hdl_variables

        processes = [self.as_hdl_HdlStmCodeBlockContainer(p) for p in processes]

        component_insts = [self.as_hdl_HdlCompInst(c)
                           for c in component_insts]
        extraVars.sort_hdl_declarations_first()
        new_m.objs = hdl_types + hdl_variables + \
            extraVars.extraVarsHdl + component_insts + processes
        return new_m

    def as_hdl_HdlModuleDef(self, o: HdlModuleDef) -> HdlModuleDef:
        # :attention: name_scope should be already set to body of module
        new_m = copy(o)
        if o.dec is not None:
            new_m.dec = self.as_hdl_HdlModuleDec(o.dec)
        return self._as_hdl_HdlModuleDef(new_m)

    def has_to_be_process(self, proc: iHdlStatement):
        raise NotImplementedError(
            "This method should be overloaded in child class")

    def can_pop_process_wrap(self, statements, hasToBeVhdlProcess):
        raise NotImplementedError(
            "This method should be overloaded in child class")

    def as_hdl_HdlStmCodeBlockContainer(self, proc: HdlStmCodeBlockContainer) -> iHdlStatement:
        """
        Serialize HdlStmCodeBlockContainer objects as process if top statement
        """
        if isinstance(proc, ALL_STATEMENT_CLASSES):
            return proc
        assert proc.parentStm is None, proc
        body = proc.statements

        hasToBeVhdlProcess = self.has_to_be_process(proc)

        with WithNameScope(self, self.name_scope.level_push(proc.name)):
            tmpVars = self.TMP_VAR_CONSTRUCTOR(self, self.name_scope)
            with TmpVarsSwap(self, tmpVars):
                statements = [self.as_hdl(s) for s in body]

                hasToBeVhdlProcess |= bool(tmpVars.extraVarsHdl)

                if hasToBeVhdlProcess:
                    tmpVars.sort_hdl_declarations_first()
                    statements = tmpVars.extraVarsHdl + statements

                if self.can_pop_process_wrap(statements, hasToBeVhdlProcess):
                    return statements[0]
                else:
                    p = HdlStmProcess()
                    p.labels.append(proc.name)

                    if not statements:
                        pass  # no body
                    elif len(statements) == 1:
                        # body made of just a singe statement
                        p.body = statements[0]
                    else:
                        p.body = HdlStmBlock()
                        assert isinstance(statements, list)
                        p.body.body = statements
                    anyIsEventDependnt = arr_any(
                        proc._sensitivity, lambda s: isinstance(s, Operator))
                    p.sensitivity = sorted([
                        self.sensitivityListItem(s, anyIsEventDependnt)
                        for s in proc._sensitivity])
                    return p



from typing import List, Optional, Union

from hdlConvertorAst.hdlAst._defs import HdlIdDef
from hdlConvertorAst.hdlAst._expr import HdlValueId
from hdlConvertorAst.hdlAst._structural import HdlModuleDec, HdlModuleDef, \
    HdlCompInst
from hwt.code import If
from hwt.doc_markers import internal
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.types.defs import BIT
from hwt.hdl.value import HValue
from hwt.serializer.utils import HdlStatement_sort_key, RtlSignal_sort_key
from hwt.synthesizer.dummyPlatform import DummyPlatform
from hwt.synthesizer.exceptions import SigLvlConfErr
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.param import Param
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal, NOT_SPECIFIED
from hwt.synthesizer.rtlLevel.rtlSyncSignal import RtlSyncSignal
from hwt.synthesizer.rtlLevel.statements_to_HdlStmCodeBlockContainers import\
    statements_to_HdlStmCodeBlockContainers


class RtlNetlist():
    """
    Hierarchical container for signals

    :ivar ~.parent: optional parent for debug and late component inspection
    :ivar ~.signals: set of all signals in this context
    :ivar ~.statements: list of all statements which are connected to signals in this context
    :ivar ~.subUnits: is set of all units in this context
    :type ~.interfaces: Dict[RtlSignal, DIRECTION]
    :ivar ~.interfaces: initialized in create_HdlModuleDef
    :type ~.ent: HdlModuleDec
    :ivar ~.ent: initialized in create_HdlModuleDec
    :type ~.arch: HdlModuleDef
    :ivar ~.arch: initialized in create_HdlModuleDef
    :ivar ~.hdl_objs: The list of HDL objects which were produced by this instance
        usually contains HdlModudeleDef but may contain imports/globals etc.
    """

    def __init__(self, parent: Optional["Unit"]=None):
        self.parent = parent
        self.signals = set()
        self.statements = set()
        self.subUnits = set()
        self.interfaces = {}
        self.hdl_objs = []
        self.ent = None
        self.arch = None
        self._port_items = []

    def sig(self, name: str, dtype=BIT, clk=None, syncRst=None,
            def_val=None, nop_val=NOT_SPECIFIED) -> Union[RtlSignal, RtlSyncSignal]:
        """
        Create new signal in this context

        :param clk: clk signal, if specified signal is synthesized
            as SyncSignal
        :param syncRst: synchronous reset signal
        :param def_val: a default value used for reset and intialization
        :param nop_val: a value which is used to drive the signal if there is no other drive
            (used to prevent latches and to specify default values for unconnected signals)
        """
        _def_val = _try_cast_any_to_HValue(def_val, dtype, True)
        if nop_val is not NOT_SPECIFIED:
            nop_val = _try_cast_any_to_HValue(nop_val, dtype, False)

        if clk is not None:
            s = RtlSyncSignal(self, name, dtype,
                              _def_val if isinstance(_def_val, HValue) else dtype.from_py(None),
                              nop_val)
            if syncRst is not None and def_val is None:
                raise SigLvlConfErr(
                    "Probably forgotten default value on sync signal %s", name)
            # dst_resolve_fn is overriden because default assign would assign to the "next" signal
            if syncRst is not None:
                r = If(syncRst._isOn(),
                       s(_def_val, dst_resolve_fn=lambda x: x)
                    ).Else(
                       s(s.next, dst_resolve_fn=lambda x: x)
                    )
            else:
                r = [
                    s(s.next, dst_resolve_fn=lambda x: x)
                ]

            if isinstance(clk, (InterfaceBase, RtlSignal)):
                clk_trigger = clk._onRisingEdge()
            else:
                # has to be tuple of (clk_sig, AllOps.RISING/FALLING_EDGE)
                clk, clk_edge = clk
                if clk_edge is AllOps.RISING_EDGE:
                    clk_trigger = clk._onRisingEdge()
                elif clk_edge is AllOps.FALLING_EDGE:
                    clk_trigger = clk._onRisingEdge()
                else:
                    raise ValueError(
                        "Invalid clock edge specification", clk_edge)

            If(clk_trigger,
               r
            )
        else:
            if syncRst:
                raise SigLvlConfErr(
                    f"Signal {name:s} has reset but has no clk")
            assert isinstance(_def_val, HValue) or (isinstance(_def_val, RtlSignal) and _def_val._const), (_def_val, "The default value needs to be constant")
            s = RtlSignal(self, name, dtype, def_val=_def_val, nop_val=nop_val)

        return s

    def create_HdlModuleDec(self, name: str,
                            store_manager: "StoreManager",
                            params: List[Param]):
        """
        Generate a module header (entity) for this module
        """
        self.ent = ent = HdlModuleDec()
        ent.name = store_manager.name_scope.checked_name(name, ent)
        ns = store_manager.hierarchy_push(ent)
        # create generics
        for p in sorted(params, key=lambda x: x._name):
            hdl_val = p.get_hdl_value()
            v = HdlIdDef()
            v.origin = p
            v.name = p.hdl_name = ns.checked_name(p._name, p)
            v.type = hdl_val._dtype
            v.value = hdl_val
            ent.params.append(v)

        return ent

    def create_HdlModuleDef(self,
                            target_platform: DummyPlatform,
                            store_manager: "StoreManager"):
        """
        Generate a module body (architecture) for this module

        * Resolve name collisions
        * Convert netlist representation to HdlProcesses
        * Remove unconnected
        * Mark visibility of signals
        """
        for proc in target_platform.beforeHdlArchGeneration:
            proc(self)

        ns = store_manager.name_scope
        mdef = HdlModuleDef()
        mdef.dec = self.ent
        mdef.module_name = HdlValueId(self.ent.name, obj=self.ent)
        mdef.name = "rtl"

        processes = sorted(self.statements, key=HdlStatement_sort_key)
        processes = sorted(statements_to_HdlStmCodeBlockContainers(processes), key=HdlStatement_sort_key)

        # add signals, variables etc. in architecture
        for s in sorted((s for s in self.signals
                        if not s.hidden and
                        s not in self.interfaces.keys()),
                        key=RtlSignal_sort_key):
                v = HdlIdDef()
                v.origin = s
                s.name = v.name = ns.checked_name(s.name, s)
                v.type = s._dtype
                v.value = s.def_val
                v.is_const = s._const
                mdef.objs.append(v)

        for p in processes:
            p.name = ns.checked_name(p.name, p)
        mdef.objs.extend(processes)
        # instantiate subUnits in architecture
        for u in self.subUnits:
            ci = HdlCompInst()
            ci.origin = u
            ci.module_name = HdlValueId(u._ctx.ent.name, obj=u._ctx.ent)
            ci.name = HdlValueId(ns.checked_name(u._name + "_inst", ci), obj=u)
            e = u._ctx.ent

            ci.param_map.extend(e.params)
            ci.port_map.extend(e.ports)

            mdef.objs.append(ci)

        self.arch = mdef
        return mdef

    def getDebugScopeName(self):
        scope = []
        p = self.parent
        while p is not None:
            scope.append(p._name)
            try:
                p = p._parent
            except AttributeError:
                break

        return ".".join(reversed(scope))


@internal
def _try_cast_any_to_HValue(v, dtype, require_const):
    if isinstance(v, RtlSignal):
        assert not require_const or v._const, \
            "Initial value of signal has to be a constant"
        return v._auto_cast(dtype)
    elif isinstance(v, HValue):
        return v._auto_cast(dtype)
    elif isinstance(v, InterfaceBase):
        return v._sig
    else:
        return dtype.from_py(v)



from copy import deepcopy
from typing import Union


class TypePath(tuple):
    """
    A path in hierarchy of structuralized type
    """
    
    def __new__ (cls, *objs):
        return super(TypePath, cls).__new__(cls, objs)

    def __truediv__(self, other: Union[int, str, "TypePath"]):
        if isinstance(other, TypePath):
            return TypePath(*self, *other)
        else:
            assert isinstance(other, (int, str)), other
            return TypePath(*self, other)

    def __copy__(self):
        return self.__class__(*self)

    def __deepcopy__(self, memo):
        res = self.__class__(*(deepcopy(x, memo) for x in self))
        memo[self] = res
        return res





class UniqList(list):
    """
    List of unique items
    """

    def __init__(self, initSeq=None):
        super(UniqList, self).__init__()
        self.__s = set()
        if initSeq is not None:
            for item in initSeq:
                self.append(item)

    def append(self, item):
        if item in self.__s:
            return False
        else:
            self.__s.add(item)
            list.append(self, item)
            return True

    def extend(self, items):
        for item in items:
            self.append(item)

    def insert(self, i, x):
        super(UniqList, self).insert(i, x)
        self.__s.add(x)

    def _get_set(self):
        return self.__s

    def intersection_set(self, other):
        return self.__s.intersection(other._get_set())

    def discard(self, item):
        if item in self.__s:
            self.remove(item)
            return True
        else:
            return False

    def remove(self, item):
        self.__s.remove(item)
        return list.remove(self, item)

    def pop(self, *args, **kwargs):
        item = list.pop(self, *args, **kwargs)
        self.__s.remove(item)
        return item

    def clear(self):
        list.clear(self)
        self.__s.clear()

    def copy(self):
        c = UniqList()
        c.extend(self)
        return c

    def __copy__(self):
        return self.copy()

    def __contains__(self, key):
        return key in self.__s



from copy import copy
from typing import Union, Dict

from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.enum import HEnum
from hwt.hdl.types.hdlType import HdlType
from hwt.hdl.types.stream import HStream
from hwt.hdl.types.struct import HStructField, HStruct
from hwt.synthesizer.typePath import TypePath

filed_filter_t = Dict[Union[int, str], "filed_filter_t"]


def HdlType_select(t: HStruct, fieldsToUse: filed_filter_t):
    """
    Select fields from type structure (rest will become padding)

    :param t: HdlType type instance
    :param fieldsToUse: dict {name:{...}} or set of names to select,
        dictionary is used to select nested fields
        in HStruct/HUnion fields/array items
        (f.e. {"struct1": {"field1", "field2"}, "field3":{}}
        will select field1 and 2 from struct1 and field3 from root)
    """

    template = []
    fieldsToUse = fieldsToUse
    foundNames = set()
    if isinstance(t, (HArray, HStream)):
        assert len(fieldsToUse) <= 1, ("select only on item 0, because it has to be same for all array items", fieldsToUse)
        k, v = list(fieldsToUse.items())[0]
        assert k == 0
        new_t = copy(t)
        new_t.elment = HdlType_select(t.element_t, v)
        return new_t
    elif isinstance(t, (Bits, HEnum)):
        # scalar
        return t
    else:
        # struct/Union
        for f in t.fields:
            name = None
            subfields = []
    
            if f.name is not None:
                try:
                    if isinstance(fieldsToUse, dict):
                        subfields = fieldsToUse[f.name]
                        name = f.name
                    else:
                        if f.name in fieldsToUse:
                            name = f.name
                except KeyError:
                    name = None
    
            if name is not None and subfields:
                new_t = HdlType_select(f.dtype, subfields)
                template.append(HStructField(new_t, name))
            else:
                template.append(HStructField(f.dtype, name))
    
            if f.name is not None:
                foundNames.add(f.name)
    
        if isinstance(fieldsToUse, dict):
            fieldsToUse = set(fieldsToUse.keys())
        assert fieldsToUse.issubset(foundNames)
    
        return t.__class__(*template)


def field_path_get_type(root: HdlType, field_path: TypePath):
    """
    Get a data type of element using field path
    """
    t = root
    for p in field_path:
        if isinstance(p, int):
            t = t.element_t
        else:
            assert isinstance(p, str), p
            t = t.field_by_name[p].dtype
    return t



from copy import copy

from hdlConvertorAst.hdlAst import HdlStmCase
from hdlConvertorAst.hdlAst._statements import HdlStmAssign
from hwt.hdl.statements.assignmentContainer import HdlAssignmentContainer
from hwt.hdl.statements.codeBlockContainer import HdlStmCodeBlockContainer
from hwt.hdl.statements.ifContainter import IfContainer
from hwt.hdl.statements.switchContainer import SwitchContainer
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import BOOL, BIT
from hwt.hdl.types.sliceVal import HSliceVal
from hwt.hdl.value import HValue
from hwt.hdl.variables import SignalItem
from hwt.serializer.exceptions import SerializerException
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class ToHdlAstVhdl2008_statements():

    def as_hdl_HdlAssignmentContainer(self, a: HdlAssignmentContainer):
        _dst = dst = a.dst
        assert isinstance(dst, SignalItem)

        if a.indexes is not None:
            for i in a.indexes:
                if isinstance(i, HSliceVal):
                    i = i.__copy__()
                dst = dst[i]

        src_t = a.src._dtype
        dst_t = dst._dtype
        correct = False
        src = a.src
        if dst_t == src_t:
            correct = True
        else:
            src = a.src
            if (isinstance(dst_t, Bits) and isinstance(src_t, Bits)):
                # std_logic <-> boolean <->  std_logic_vector(0 downto 0) auto conversions
                while not (dst_t == src_t):
                    # while is used because the casting could be required multiple times
                    correct = False
                    if dst_t.bit_length() == src_t.bit_length() == 1:
                        if dst_t.force_vector and not src_t.force_vector:
                            dst = dst[0]
                            correct = True
                        elif not dst_t.force_vector and src_t.force_vector:
                            src = src[0]
                            correct = True
                        elif src_t == BOOL:
                            src = src._ternary(BIT.from_py(1), BIT.from_py(0))
                            correct = True
                    elif not src_t.strict_width:
                        if isinstance(src, HValue):
                            src = copy(src)
                            if a.indexes:
                                raise NotImplementedError()

                            src._dtype = dst_t
                            correct = True
                        else:
                            raise NotImplementedError()
                            pass

                    src_t = src._dtype
                    dst_t = dst._dtype

                    if not correct:
                        # automatic type cast can not be performed
                        break
        if correct:
            src = self.as_hdl(src)
            hdl_a = HdlStmAssign(src, self.as_hdl(dst))
            hdl_a.is_blocking = _dst.virtual_only
            return hdl_a

        raise SerializerException(
            f"{dst} = {a.src}  is not valid assignment\n"
            f" because types are different ({dst._dtype}; {a.src._dtype})")

    def as_hdl_SwitchContainer(self, sw: SwitchContainer) -> HdlStmCase:
        s = HdlStmCase()
        switchOn = sw.switchOn
        if isinstance(switchOn, RtlSignalBase) and switchOn.hidden:
            _, switchOn = self.tmpVars.create_var_cached("tmpTypeConv_", switchOn._dtype, def_val=switchOn)

        s.switch_on = self.as_hdl_cond(switchOn, False)
        s.cases = cases = []
        for key, statements in sw.cases:
            key = self.as_hdl_Value(key)
            cases.append((key, self.as_hdl_statements(statements)))

        s.default = self.as_hdl_statements(sw.default)
        return s

    def can_pop_process_wrap(self, stms, hasToBeVhdlProcess):
        if hasToBeVhdlProcess or len(stms) > 1:
            return False
        else:
            assert len(stms) == 1, stms
            return True

    def has_to_be_process(self, proc: HdlStmCodeBlockContainer):
        for x in proc.statements:
            if isinstance(x, (IfContainer, SwitchContainer)):
                return True
        return False



from hwt.pyUtils.arrayQuery import flatten
from hwt.synthesizer.interface import Interface
from hwtSimApi.agents.base import AgentBase
from hwtSimApi.hdlSimulator import HdlSimulator
from ipCorePackager.constants import INTF_DIRECTION


class UniversalCompositeAgent(AgentBase):
    """
    Composite agent which just instanciates agents for every subinterface
    """

    def __init__(self, sim: HdlSimulator, intf: Interface):
        self.__enable = True
        super(UniversalCompositeAgent, self).__init__(sim, intf)
        for i in intf._interfaces:
            i._initSimAgent(sim)

    def getEnable(self):
        return self.__enable

    def setEnable(self, v: bool):
        """
        Distribute change of enable on child agents
        """
        self.__enable = v
        for o in self.intf._interfaces:
            o._ag.setEnable(v)

    def getDrivers(self):
        return list(flatten(
            (i._ag.getMonitors() if i._direction == INTF_DIRECTION.MASTER else i._ag.getDrivers()
            for i in self.intf._interfaces),
            level=1
        ))

    def getMonitors(self):
        return list(flatten(
            (i._ag.getMonitors() if i._direction == INTF_DIRECTION.MASTER else i._ag.getDrivers()
            for i in self.intf._interfaces),
            level=1
        ))



from hwt.interfaces.agents.structIntf import StructIntfAgent


class UnionSourceAgent(StructIntfAgent):

    def getMonitors(self):
        sel = self.intf._select
        for intf in self.intf._interfaces:
            if intf is sel:
                yield from intf._ag.getDrivers()
            else:
                yield from intf._ag.getMonitors()

    def getDrivers(self):
        sel = self.intf._select
        for intf in self.intf._interfaces:
            if intf is sel:
                yield from intf._ag.getMonitors()
            else:
                yield from intf._ag.getDrivers()



from hwt.doc_markers import internal
from hwt.hdl.types.hdlType import HdlType


class HSlice(HdlType):
    """
    Slice type, used for selecting items from arrays or vectors
    """

    @internal
    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdl.types.sliceVal import HSliceVal
            cls._valCls = HSliceVal
            return cls._valCls



from typing import Union

from hwt.doc_markers import internal
from hwt.hdl.types.defs import BIT
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from ipCorePackager.constants import DIRECTION
from hwt.hdl.value import HValue


def rename_signal(unit_instance: "Unit",
                  sig: Union[RtlSignalBase, int, bool],
                  name: str):
    """
    Wrap signal or value in signal of specified name

    :attention: output signal is driven by new signal of a specified name
        this means that the assigning to a new signal does not drive a original signal
    """

    if isinstance(sig, (int, bool)):
        t = BIT
    else:
        t = sig._dtype
    if isinstance(sig, (HValue, int, bool)):
        s = unit_instance._sig(name, t, def_val=sig)
    else:
        s = unit_instance._sig(name, t)
        s(sig)
    return s


def connect_optional(src: InterfaceBase, dst: InterfaceBase,
                     check_fn=lambda intf_a, intf_b: (True, [])):
    """
    Connect interfaces and ignore all missing things

    :param check_fn: filter function(intf_a, intf_b) which check if interfaces should be connected
        returns tuple (do_check, extra_connection_list)
    """
    return list(_connect_optional(src, dst, check_fn, False))


@internal
def _connect_optional(src: InterfaceBase, dst: InterfaceBase, check_fn, dir_reverse):
    do_connect, extra_connections = check_fn(src, dst)
    yield from extra_connections
    if not do_connect:
        return

    if not src._interfaces:
        assert not dst._interfaces, (src, dst)
        if dir_reverse:
            yield src(dst)
        else:
            yield dst(src)

    for _s in src._interfaces:
        _d = getattr(dst, _s._name, None)
        if _d is None:
            # if the interfaces does not have subinterface of same name
            continue

        if _d._masterDir == DIRECTION.IN:
            rev = not dir_reverse
        else:
            rev = dir_reverse

        yield from _connect_optional(_s, _d, check_fn, rev)


@internal
def _intfToSig(obj):
    if isinstance(obj, InterfaceBase):
        return obj._sig
    else:
        return obj


@internal
def _mkOp(fn):
    """
    Function to create variadic operator function

    :param fn: function to perform binary operation
    """

    def op(*operands, key=None) -> RtlSignalBase:
        """
        :param operands: variadic parameter of input uperands
        :param key: optional function applied on every operand
            before processing
        """
        assert operands, operands
        top = None
        if key is not None:
            operands = map(key, operands)

        for s in operands:
            if top is None:
                top = s
            else:
                top = fn(top, s)
        return top

    return op




from itertools import islice
from typing import Dict, List, Tuple, Union, Optional

from hwt.code import Concat
from hwt.doc_markers import internal
from hwt.hdl.operator import isConst
from hwt.hdl.statements.assignmentContainer import HdlAssignmentContainer
from hwt.hdl.statements.codeBlockContainer import HdlStmCodeBlockContainer
from hwt.hdl.statements.ifContainter import IfContainer
from hwt.hdl.statements.statement import HdlStatement
from hwt.hdl.statements.switchContainer import SwitchContainer
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.bitsVal import BitsVal
from hwt.hdl.types.defs import SLICE
from hwt.hdl.types.sliceVal import HSliceVal
from hwt.hdl.value import HValue
from hwt.pyUtils.uniqList import UniqList
from hwt.serializer.utils import RtlSignal_sort_key, HdlStatement_sort_key
from hwt.synthesizer.rtlLevel.constants import NOT_SPECIFIED
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal


def find_independent_slice_drivers(stm: HdlStatement):
    if isinstance(stm, HdlAssignmentContainer):
        if stm.indexes and len(stm.indexes) == 1 and isinstance(stm.dst._dtype, Bits):
            dst = stm.dst
            for i in stm.indexes:
                if not isConst(i):
                    return

            can_directly_replace_with_src_expr = stm.parentStm is None
            yield (
                dst,
                tuple(stm.indexes),
                can_directly_replace_with_src_expr,
                stm.src if can_directly_replace_with_src_expr else None
            )
    else:
        for _stm in stm._iter_stms():
            yield from find_independent_slice_drivers(_stm)


def _format_indexes(indexes):
    return tuple(
        (int(i) + 1, int(i))
            if isinstance(i, BitsVal) else
        (int(i.val.start), int(i.val.stop))
        for i in indexes)


def extract_part_drivers_stm(stm: HdlStatement,
                             signal_parts: Dict[RtlSignal,
                                                List[Tuple[RtlSignal, List[HValue]]]]
                             ) -> bool:
    """
    :returns: True if statement was modified
    """
    if isinstance(stm, HdlAssignmentContainer):
        dst = stm.dst
        parts = signal_parts.get(dst, None)
        if parts is None:
            return False
        if stm.indexes and len(stm.indexes) == 1:
            indexes = _format_indexes(stm.indexes)
            new_dsts, do_remove_stm = parts[indexes]
        else:
            # collect only parts which do not have sub parts (are primitive parts)
            new_dsts = []
            for k, d in parts.items():
                if not isinstance(d, list):
                    new_dsts.append(k)
            new_dsts.sort()
            do_remove_stm = False

        if isinstance(new_dsts, list):
            if stm.parentStm is None:
                return False
            assert len(new_dsts) > 1, (dst, new_dsts, stm)
            # assert not do_remove_stm, (dst, new_dsts, stm)
            # the driven slice was split to multiple sub slices
            replacement = []
            dst_offset = new_dsts[0][-1][1]
            for i in new_dsts:
                new_dst = parts[i][0]
                new_src = stm.src
                for _i in i:
                    high, low = _i[0] - dst_offset, _i[1] - dst_offset
                    assert high > 0 and low >= 0, dst_offset
                    assert high > low, (dst, stm, (high, low))
                    new_src = new_src[high:low]
                a = new_dst(new_src)
                replacement.append(a)

            # it has to have parent statement because it needs to be nested
            # because otherwise it would not have some overlapping parts driven diferently
            # under some condition
            stm.parentStm._replace_child_statement(stm, replacement, False)
            if do_remove_stm:
                stm._destroy()

        elif do_remove_stm:
            # remove current assignment because we are using src directly
            # assert stm.parentStm is None, (stm, stm.parentStm)
            stm._destroy()
        else:
            # rewrite the HdlAssignmentContainer instance to use new dst
            replacement = [new_dsts(stm.src), ]
            stm.parentStm._replace_child_statement(stm, replacement, False)
        return True

    elif isinstance(stm, (IfContainer, SwitchContainer, HdlStmCodeBlockContainer)):
        modified = False
        for _stm in stm._iter_stms():
            modified |= extract_part_drivers_stm(_stm, signal_parts)
        if modified:
            assert not stm._enclosed_for, "_enclosed_for is expected not to be initialized yet"
            outputs = stm._outputs
            inputs = stm._inputs
            stm._outputs = UniqList()
            stm._inputs = UniqList()
            stm._collect_io()
            if stm.parentStm is None:
                for o in outputs:
                    if o not in stm._outputs:
                        o.drivers.remove(stm)

                for i in inputs:
                    if i not in stm._inputs:
                        i.endpoints.remove(stm)

            return True

    else:
        raise NotImplementedError("Unknown statement ", stm)

    return False


@internal
def construct_tmp_dst_sig_for_slice(dst: RtlSignal,
                                    indexes: List[Union[BitsVal, HSliceVal]],
                                    src: Optional[RtlSignal],
                                    is_signal_needed: bool) -> RtlSignal:
    """
    Construct a tmp signal or value which will be used instead of slice from original signal

    :param dst: a signal which slice we want to generate tmp signal for
    :param indexes: a indexes to specify the slice of the dst
    :param is_signal_needed: True if we need a signal which will we drive later, else returns HValue instance
        resolved from default and nop value
    """
    if is_signal_needed:
        name = dst.name

    def_val = dst.def_val
    nop_val = dst._nop_val
    for i in indexes:
        def_val = def_val[i]
        if nop_val is not NOT_SPECIFIED:
            nop_val = nop_val[i]

        if is_signal_needed:
            dst = dst[i]
            if isinstance(i, HSliceVal):
                if int(i.val.step) == -1:
                    stop = int(i.val.stop)
                    start = int(i.val.start)
                    name = f"{name}_{start - 1:d}downto{stop:d}"
                else:
                    raise NotImplementedError(i.val.step)
            else:
                _i = int(i)
                name = f"{name:s}_{_i:d}"

    if is_signal_needed:
        tmp_sig = dst.ctx.sig(name, dst._dtype, def_val=def_val, nop_val=nop_val)
        return tmp_sig
    elif src is not None:
        return src
    elif nop_val is not NOT_SPECIFIED:
        return nop_val
    else:
        return def_val


def resolve_splitpoints(s: RtlSignal, parts):
    split_points = set()
    add_split_point = split_points.add
    for i, _, _ in parts:
        if len(i) != 1:
            raise NotImplementedError(s, i)
        i = i[0]

        if isinstance(i, BitsVal):
            # index is normal integer
            i = int(i)
            add_split_point(i)
            add_split_point(i + 1)
        else:
            # index is slice
            assert isinstance(i, HSliceVal), (s, i)
            add_split_point(int(i.val.start))
            add_split_point(int(i.val.stop))

    if isinstance(s._dtype, Bits):
        # add boundary points in the case something is unconnected
        add_split_point(0)
        add_split_point(s._dtype.bit_length())
    else:
        raise NotImplementedError(s._dtype)

    return split_points


def resolve_final_parts_from_splitpoints_and_parts(signal_parts):
    # :type: Dict[RtlSignal, Dict[Tuple[Tuple[int, int], ...], Union[HValue, RtlSignal]]]
    final_signal_parts = {}
    # split part intervals to non-overlapping chunks
    for s, parts in sorted(signal_parts.items(), key=lambda x: RtlSignal_sort_key(x[0])):
        split_point = resolve_splitpoints(s, parts)
        split_point = sorted(split_point)
        # prepare part signals
        new_parts = []
        new_parts_dict = {}
        split_i = 0
        end = 0
        # :attention: parts are likely to contain parts with same indexes
        for indexes, can_directly_replace_with_src_expr, src in sorted(parts, key=lambda x: x[0]):
            if len(indexes) != 1:
                raise NotImplementedError()
            i = indexes[0]
            split_p = split_point[split_i]
            if isinstance(i, BitsVal):
                low = int(i)
                high = low + 1
                index_key = ((high, low),)
            else:
                assert isinstance(i, HSliceVal), (s, i)
                if i.val.step != -1:
                    raise NotImplementedError(s, i)
                high, low = int(i.val.start), int(i.val.stop)
                index_key = ((high, low),)

            while split_p < low:
                # some parts at the beginning are skiped
                # that means that that part is not driven by anything
                # and we need to check default and nop value
                part_indexes = (SLICE.from_py(slice(low, split_p , -1)),)
                _src = construct_tmp_dst_sig_for_slice(s, part_indexes, None, isinstance(s._nop_val, RtlSignal))
                new_parts.append(_src)
                _index_key = ((low, split_p),)
                new_parts_dict[_index_key] = _src, True
                split_i += 1
                split_p = split_point[split_i]

            this_start_split_p_i = split_i
            if split_p > low:
                # some parts at the beginning were already resolved
                # This can happen if there was some part which started on some <= index and overlaps with this part.
                try:
                    _, _can_directly_replace_with_src_expr = new_parts_dict[index_key]
                    assert not _can_directly_replace_with_src_expr, (s, index_key)
                    # was already resolved and checked no need to check it again
                    continue
                except KeyError:
                    pass

                for i in range(split_i, -1, -1):
                    _sp = split_point[i]
                    if _sp == low:
                        this_start_split_p_i = i

            assert split_point[this_start_split_p_i] == low
            # just at the start of this slice
            next_split_p = split_point[this_start_split_p_i + 1]
            assert next_split_p <= high, "The next split point can be at most end of current part"
            if next_split_p == high:
                assert this_start_split_p_i == split_i, "We should see this part for the first time or the split_i should already be higher"
                # all bits on this slice are alwyas driven at once, we can instantiate whole part
                assert split_p == low
                _src = construct_tmp_dst_sig_for_slice(s, indexes, src, not can_directly_replace_with_src_expr)
                new_parts.append(_src)
                assert index_key not in new_parts_dict, (s, index_key)
                new_parts_dict[index_key] = _src, can_directly_replace_with_src_expr
                split_i += 1
            else:
                # list of part keys for later search
                _split_parts = []
                prev_sp = split_point[this_start_split_p_i]
                dst_offset = low
                assert not can_directly_replace_with_src_expr
                # continue instanciating parts until we reach the end of this part
                for sp_i, sp in zip(range(this_start_split_p_i + 1, len(split_point)),
                                    islice(split_point, this_start_split_p_i + 1, None)):
                    # need to generate sub slice
                    # because this slice has actually multiple individualy driven parts

                    # we need to generate all slice parts because there could be a case where only some sub parts are
                    # driven elsewhere and we would othervise resolve those segments as a constantly driven
                    # but they are in fact driven from this slice
                    if sp > high:
                        break

                    part_key = ((sp, prev_sp),)
                    if sp_i <= split_i:
                        # check if the slice is not driven from some top level constant assignment
                        # which would result is multiple drivers of this slice
                        assert src is None
                        existing_part, _can_directly_replace_with_src_expr = new_parts_dict[part_key]
                        assert not _can_directly_replace_with_src_expr, (s, low, high, existing_part)
                        assert not can_directly_replace_with_src_expr, (s, low, high, existing_part)
                        assert isinstance(existing_part, RtlSignal), (s, low, high, existing_part)
                    else:
                        assert sp_i == split_i + 1, (s, sp_i, split_i)
                        # get actual input signal
                        if src is None:
                            _src = None
                        else:
                            _src = src[sp - dst_offset:prev_sp - dst_offset]

                        part_indexes = (SLICE.from_py(slice(sp, prev_sp, -1)),)
                        _src = construct_tmp_dst_sig_for_slice(s, part_indexes, _src, True)

                        new_parts.append(_src)
                        new_parts_dict[part_key] = _src, can_directly_replace_with_src_expr

                        split_i += 1
                    _split_parts.append(part_key)
                    prev_sp = sp

                new_parts_dict[index_key] = _split_parts, False

            end = max(end, high)

        if end < split_point[-1]:
            # something unconnected at the end
            high, low = split_point[-1], end
            part_indexes = (SLICE.from_py(slice(high, low , -1)),)
            _src = construct_tmp_dst_sig_for_slice(s, part_indexes, None, isinstance(s._nop_val, RtlSignal))
            new_parts.append(_src)
            index_key = ((high, low),)
            new_parts_dict[index_key] = _src, True

        # construct assignment of concatenation from all parts
        assert new_parts, (s, parts)
        s(Concat(*reversed(new_parts)))
        final_signal_parts[s] = new_parts_dict

    return final_signal_parts


@internal
def extract_part_drivers(netlist: "RtlNetlist"):
    """
    .. code-block:: verilog

        if (c0)
            s[0] <= x;
        if (c1)
            s[1] <= y;

    to

    .. code-block:: verilog

        wire s_0_tmp;
        wire s_1_tmp;
        assign s <= {s_1_tmp, s_0_tmp};
        if (c0)
            s_0_tmp <= x;
        if (c1)
            s_1_tmp <= y;

    """
    signal_parts = {}
    for stm in sorted(netlist.statements, key=HdlStatement_sort_key):
        for s, indexes, can_directly_replace_with_src_expr, src in find_independent_slice_drivers(stm):
            signal_parts.setdefault(s, []).append((indexes, can_directly_replace_with_src_expr, src))

    if not signal_parts:
        return

    final_signal_parts = resolve_final_parts_from_splitpoints_and_parts(signal_parts)
    for stm in sorted(netlist.statements, key=HdlStatement_sort_key):
        extract_part_drivers_stm(stm, final_signal_parts)



from typing import Union

from hdlConvertorAst.hdlAst._expr import HdlValueId, HdlOp, HdlOpType
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_call
from hdlConvertorAst.translate.common.name_scope import LanguageKeyword
from hwt.code import If
from hwt.doc_markers import internal
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.statements.assignmentContainer import HdlAssignmentContainer
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import BOOL, INT
from hwt.hdl.value import HValue
from hwt.serializer.hwt.ops import ToHdlAstHwt_ops
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal


@internal
def isResultOfTypeConversion(sig):
    if len(sig.drivers) != 1:
        return False

    if sig.hidden:
        d = sig.singleDriver()
        return d.operator != AllOps.INDEX

    return False


class ToHdlAstVhdl2008_ops():
    op_transl_dict = {
        **ToHdlAstHwt_ops.op_transl_dict,
        AllOps.RISING_EDGE: HdlOpType.RISING,
        AllOps.FALLING_EDGE: HdlOpType.FALLING,
    }
    _cast_ops = {
        AllOps.BitsAsSigned: "SIGNED",
        AllOps.BitsAsUnsigned: "UNSIGNED",
        AllOps.BitsAsVec: "STD_LOGIC_VECTOR",
    }

    @internal
    def _tmp_var_for_ternary(self, val: RtlSignal):
        """
        Optionally convert boolean to std_logic_vector
        """
        isNew, o = self.tmpVars.create_var_cached(
            "tmpTernary_",
            val._dtype,
            postponed_init=True,
            extra_args=(val, bool, 1, 0))
        if isNew:
            cond, ifTrue, ifFalse = val.drivers[0].operands
            if_ = If(cond)
            if_.ifTrue.append(HdlAssignmentContainer(ifTrue, o,
                                         virtual_only=True,
                                         parentStm=if_))
            if_.ifFalse = []
            if_.ifFalse.append(HdlAssignmentContainer(ifFalse, o,
                                          virtual_only=True,
                                          parentStm=if_))
            if_._outputs.append(o)
            for obj in (cond, ifTrue, ifFalse):
                if isinstance(obj, RtlSignalBase):
                    if_._inputs.append(obj)
            o.drivers.append(if_)
            self.tmpVars.finish_var_init(o)

        return o

    def _as_Bits(self, val: Union[RtlSignal, HValue]):
        if val._dtype == BOOL:
            bit1_t = Bits(1)
            isNew, o = self.tmpVars.create_var_cached(
                "tmpBool2std_logic_",
                bit1_t,
                postponed_init=True,
                extra_args=(val, int, 1, 0))
            if isNew:
                ifTrue, ifFalse = bit1_t.from_py(1), bit1_t.from_py(0)
                if_ = If(val)
                if_.ifTrue.append(HdlAssignmentContainer(ifTrue, o, virtual_only=True, parentStm=if_))
                if_.ifFalse = []
                if_.ifFalse.append(HdlAssignmentContainer(ifFalse, o, virtual_only=True, parentStm=if_))
                if_._outputs.append(o)
                o.drivers.append(if_)
                self.tmpVars.finish_var_init(o)
            return o
        else:
            assert isinstance(val._dtype, Bits), val._dtype
            return val

    def _as_Bits_vec(self, val: Union[RtlSignal, HValue]):
        val = self._as_Bits(val)
        t = val._dtype
        if not t.force_vector and t.bit_length() == 1:
            # std_logic -> std_logic_vector
            std_logic_vector = Bits(1, signed=t.signed, force_vector=True)
            isNew, o = self.tmpVars.create_var_cached(
                "tmp_std_logic2vector_",
                std_logic_vector,
                postponed_init=True,
                extra_args=(val, std_logic_vector))
            if isNew:
                o.drivers.append(HdlAssignmentContainer(val, o, virtual_only=True))
                self.tmpVars.finish_var_init(o)
            return o
        else:
            # already a std_logic_vector
            return val

    def as_hdl_operand(self, operand: Union[RtlSignal, HValue]):
        # no nested ternary in expressions like
        # ( '1'  WHEN r = f ELSE  '0' ) & "0"
        # extract them as a tmp variable
        try:
            isTernaryOp = operand.hidden\
                and operand.drivers[0].operator == AllOps.TERNARY
        except (AttributeError, IndexError):
            isTernaryOp = False

        if isTernaryOp:
            # rewrite ternary operator as if
            operand = self._tmp_var_for_ternary(operand)
        return self.as_hdl(operand)

    def apply_cast(self, t_name, op):
        return hdl_call(HdlValueId(t_name, obj=LanguageKeyword()),
                        [op, ])

    def as_hdl_Operator(self, op: Operator):
        ops = op.operands
        o = op.operator

        if o == AllOps.INDEX:
            op0, op1 = ops
            if isinstance(op0, RtlSignalBase) and isResultOfTypeConversion(op0):
                _, op0 = self.tmpVars.create_var_cached("tmpTypeConv_", op0._dtype, def_val=op0)

            # if the op0 is not signal or other index index operator it is extracted
            # as tmp variable
            op0 = self.as_hdl_operand(op0)

            if isinstance(op1._dtype, Bits) and op1._dtype != INT:
                if op1._dtype.signed is None:
                    if op1._dtype.bit_length() == 1 and not op1._dtype.force_vector:
                        _, op1 = self.tmpVars.create_var_cached("tmpTypeConv_", Bits(1, force_vector=True), def_val=op1)
                    _op1 = self.as_hdl_operand(op1)
                    _op1 = self.apply_cast("UNSIGNED", _op1)
                else:
                    _op1 = self.as_hdl_operand(op1)

                _op1 = self.apply_cast("TO_INTEGER", _op1)
            else:
                _op1 = self.as_hdl_operand(op1)

            return HdlOp(HdlOpType.INDEX, [op0, _op1])
        elif o == AllOps.TERNARY:
            _c, _op0, _op1 = ops
            op0 = self.as_hdl_cond(_c, True)
            op1 = self.as_hdl_operand(_op0)
            t0 = _op0._dtype
            t1 = _op1._dtype
            if not (t0 == t1):
                assert isinstance(t0, Bits) and\
                       isinstance(t1, Bits) and\
                       t0.bit_length() == t1.bit_length() and\
                       bool(t0.signed) == bool(t1.signed), (t0, t1)
                _, _op1 = self.tmpVars.create_var_cached("tmpTypeConv_", t0, def_val=_op1)

            op2 = self.as_hdl_operand(_op1)
            return HdlOp(HdlOpType.TERNARY, [op0, op1, op2])
        else:
            _o = self._cast_ops.get(o, None)
            if _o is not None:
                op0 = ops[0]
                op0 = self._as_Bits_vec(op0)
                if isinstance(op0, RtlSignalBase) and op0.hidden:
                    _, op0 = self.tmpVars.create_var_cached("tmpTypeConv_", op0._dtype, def_val=op0)
                return self.apply_cast(_o, self.as_hdl_operand(op0))

            o = self.op_transl_dict[o]
            if len(ops) == 2:
                res_t = op.result._dtype
                op0, op1 = ops
                if isinstance(res_t, Bits) and res_t != BOOL:
                    op0 = self._as_Bits(op0)
                    op1 = self._as_Bits(op1)
                op0 = self.as_hdl_operand(op0)
                op1 = self.as_hdl_operand(op1)
                return HdlOp(o, [op0, op1])
            return HdlOp(o, [self.as_hdl_operand(o2)
                             for o2 in ops])



from typing import TypeVar, Generic, Union, Optional, Dict

from hwt.hdl.constants import DIRECTION
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import BIT, BIT_N
from hwt.hdl.types.hdlType import HdlType
from hwt.interfaces.agents.bramPort import BramPortAgent
from hwt.interfaces.agents.bramPort import BramPort_withoutClkAgent
from hwt.interfaces.agents.fifo import FifoReaderAgent
from hwt.interfaces.agents.fifo import FifoWriterAgent
from hwt.interfaces.agents.handshaked import HandshakeSyncAgent
from hwt.interfaces.agents.handshaked import HandshakedAgent
from hwt.interfaces.agents.rdSynced import RdSyncedAgent
from hwt.interfaces.agents.regCntrl import RegCntrlAgent
from hwt.interfaces.agents.signal import SignalAgent
from hwt.interfaces.agents.vldSynced import VldSyncedAgent
from hwt.interfaces.signalOps import SignalOps
from hwt.synthesizer.interface import Interface
from hwt.synthesizer.param import Param
from hwtSimApi.agents.clk import ClockAgent
from hwtSimApi.agents.rst import PullDownAgent
from hwtSimApi.agents.rst import PullUpAgent
from hwtSimApi.hdlSimulator import HdlSimulator
from hwtSimApi.utils import freq_to_period


D = DIRECTION

T = TypeVar("T", bound=HdlType)


class Signal(SignalOps, Interface, Generic[T]):
    """
    Basic wire interface

    :ivar ~._dtype: type of signal
    :ivar ~._sig: RtlSignal instance (physical representation of this logical signal)
    :ivar ~._sigInside: _sig after to_rtl conversion is made
        (after to_rtl conversion _sig is signal for parent unit
        and _sigInside is signal in original unit, this separates process
        of translating units)
    :note: _sigInside is None if the body of component was not elaborated yet
    :ivar _isAccessible: flag which is set False if the signal is inside of some elaborated unit
    """

    def __init__(self,
                 dtype: HdlType=BIT,
                 masterDir: DIRECTION=D.OUT,
                 hdl_name:Optional[Union[str, Dict[str, str]]]=None,
                 loadConfig: bool=True):
        self._sig = None
        self._sigInside = None
        self._isAccessible = True
        super().__init__(masterDir=masterDir,
                         hdl_name=hdl_name,
                         loadConfig=loadConfig)
        self._dtype = dtype

    def _clean(self, lockNonExternal=True):
        """
        :see: :func:`Interface._clean`
        """
        self._sigInside = self._sig
        self._sig = None
        if lockNonExternal and not self._isExtern:
            self._isAccessible = False
        if self._interfaces:
            Interface._clean(self, lockNonExternal=lockNonExternal)

    def __copy__(self):
        """
        Create new instance of interface of same type and configuration
        """
        intf = self.__class__(masterDir=self._masterDir,
                              dtype=self._dtype)
        intf._updateParamsFrom(self)
        return intf

    def _initSimAgent(self, sim: HdlSimulator):
        self._ag = SignalAgent(sim, self)


def VectSignal(width: int,
               signed: Union[bool, None]=None,
               masterDir=D.OUT,
               hdl_name:Optional[Union[str, Dict[str, str]]]=None,
               loadConfig=True):
    """
    Create basic :class:`.Signal` interface where type is vector
    """
    return Signal(Bits(width, signed, force_vector=True),
                  masterDir,
                  hdl_name,
                  loadConfig)


class Clk(Signal):
    """
    Basic :class:`.Signal` interface which is interpreted as clock signal
    """
    DEFAULT_FREQ = int(100e6)

    def _config(self):
        self.FREQ = Param(self.DEFAULT_FREQ)

    def _getIpCoreIntfClass(self):
        from hwt.interfaces.std_ip_defs import IP_Clk
        return IP_Clk

    def _initSimAgent(self, sim: HdlSimulator):
        self._ag = ClockAgent(sim, self, period=int(freq_to_period(self.FREQ)))


class Rst(Signal[Bits]):
    """
    Basic :class:`.Signal` interface which is interpreted as reset signal
    """

    def _getIpCoreIntfClass(self):
        from hwt.interfaces.std_ip_defs import IP_Rst
        return IP_Rst

    def _initSimAgent(self, sim: HdlSimulator):
        clk = self._getAssociatedClk()
        self._ag = PullDownAgent(sim, self,
                                 initDelay=int(0.6 * freq_to_period(clk.FREQ)))


class Rst_n(Signal[Bits]):
    """
    Basic :class:`.Signal` interface which is interpreted as reset signal
    with negative polarity (active in 0)
    """

    def __init__(self,
                 masterDir=D.OUT,
                 dtype=BIT_N,
                 hdl_name:Optional[Union[str, Dict[str, str]]]=None,
                 loadConfig=True):
        super(Rst_n, self).__init__(masterDir=masterDir,
                                    dtype=dtype,
                                    hdl_name=hdl_name,
                                    loadConfig=loadConfig)

    def _getIpCoreIntfClass(self):
        from hwt.interfaces.std_ip_defs import IP_Rst_n
        return IP_Rst_n

    def _initSimAgent(self, sim: HdlSimulator):
        clk = self._getAssociatedClk()
        self._ag = PullUpAgent(sim, self,
                               initDelay=int(0.6 * freq_to_period(clk.FREQ)))


class VldSynced(Interface):
    """
    Interface data+valid signal, if vld=1 then data are valid and slave should
    accept them
    """

    def _config(self):
        self.DATA_WIDTH = Param(64)

    def _declr(self):
        self.data = VectSignal(self.DATA_WIDTH)
        self.vld = Signal()

    def _initSimAgent(self, sim: HdlSimulator):
        self._ag = VldSyncedAgent(sim, self)


class RdSynced(Interface):
    """
    Interface data+ready signal, if rd=1 then slave has read data and master
    should actualize data
    """

    def _config(self):
        self.DATA_WIDTH = Param(64)

    def _declr(self):
        self.data = VectSignal(self.DATA_WIDTH)
        self.rd = Signal(masterDir=D.IN)

    def _initSimAgent(self, sim: HdlSimulator):
        self._ag = RdSyncedAgent(sim, self)


class Handshaked(VldSynced):
    """
    Interface data+ready+valid signal, if rd=1 slave is ready to accept data,
    if vld=1 master is sending data,
    if rd=1 and vld=1 then data is transfered otherwise master
    and slave has to wait on each other

    :attention: one rd/vld is set it must not go down until transaction is made
    """

    def _declr(self):
        super()._declr()
        self.rd = Signal(masterDir=D.IN)

    def _initSimAgent(self, sim: HdlSimulator):
        self._ag = HandshakedAgent(sim, self)


class HandshakeSync(Interface):
    """
    Only synchronization interface, like vld+rd signal with meaning
    like in :class:`.Handshaked` interface

    :ivar ~.rd: when high slave is ready to receive data
    :ivar ~.vld: when high master is sending data to slave

    transaction happens when both ready and valid are high
    """

    def _declr(self):
        self.vld = Signal()
        self.rd = Signal(masterDir=D.IN)

    def _initSimAgent(self, sim: HdlSimulator):
        self._ag = HandshakeSyncAgent(sim, self)


class ReqDoneSync(Interface):
    """
    Synchronization interface, if req=1 slave begins operation and when
    it's done it asserts done=1 for one clk tick req does not need to stay high
    """

    def _declr(self):
        self.req = Signal()
        self.done = Signal(masterDir=D.IN)


class BramPort_withoutClk(Interface):
    """
    Basic BRAM port
    """

    def _config(self):
        self.ADDR_WIDTH = Param(32)
        self.DATA_WIDTH = Param(64)
        self.HAS_R = Param(True)
        self.HAS_W = Param(True)
        self.HAS_BE = Param(False)

    def _declr(self):
        assert self.HAS_R or self.HAS_W, "has to have at least read or write part"

        self.addr = VectSignal(self.ADDR_WIDTH)
        DATA_WIDTH = self.DATA_WIDTH
        if self.HAS_W:
            self.din = VectSignal(DATA_WIDTH)

        if self.HAS_R:
            self.dout = VectSignal(DATA_WIDTH, masterDir=D.IN)

        self.en = Signal()
        if (self.HAS_R and self.HAS_W) or (self.HAS_W and self.HAS_BE):
            # in write only mode we do not need this as well as we can use "en"
            if self.HAS_BE:
                assert DATA_WIDTH % 8 == 0, DATA_WIDTH
                self.we = VectSignal(DATA_WIDTH // 8)
            else:
                self.we = Signal()

    def _getWordAddrStep(self):
        """
        :return: size of one word in unit of address
        """
        return 1

    def _getAddrStep(self):
        """
        :return: how many bits is one unit of address (e.g. 8 bits for
            char * pointer, 36 for 36 bit bram)
        """
        return int(self.DATA_WIDTH)

    def _getIpCoreIntfClass(self):
        from hwt.interfaces.std_ip_defs import IP_BlockRamPort
        return IP_BlockRamPort

    def _initSimAgent(self, sim: HdlSimulator):
        self._ag = BramPort_withoutClkAgent(sim, self)


class BramPort(BramPort_withoutClk):
    """
    BRAM port with it's own clk
    """

    def _declr(self):
        self.clk = Signal(masterDir=D.OUT)
        with self._associated(clk=self.clk):
            super()._declr()

        self._make_association(clk=self.clk)

    def _initSimAgent(self, sim: HdlSimulator):
        self._ag = BramPortAgent(sim, self)


class FifoWriter(Interface):
    """
    FIFO write port interface
    """

    def _config(self):
        self.DATA_WIDTH = Param(8)

    def _declr(self):
        self.en = Signal()
        self.wait = Signal(masterDir=D.IN)
        if self.DATA_WIDTH:
            self.data = VectSignal(self.DATA_WIDTH)

    def _initSimAgent(self, sim: HdlSimulator):
        self._ag = FifoWriterAgent(sim, self)

    def _getIpCoreIntfClass(self):
        from hwt.interfaces.std_ip_defs import IP_FifoWriter
        return IP_FifoWriter


class FifoReader(Interface):
    """
    FIFO read port interface
    """

    def _config(self):
        FifoWriter._config(self)

    def _declr(self):
        FifoWriter._declr(self)
        self.en._masterDir = D.IN
        self.wait._masterDir = D.OUT

    def _initSimAgent(self, sim: HdlSimulator):
        self._ag = FifoReaderAgent(sim, self)

    def _getIpCoreIntfClass(self):
        from hwt.interfaces.std_ip_defs import IP_FifoReader
        return IP_FifoReader


class RegCntrl(Interface):
    """
    Register control interface, :class:`.Signal` for read, :class:`.VldSynced`
    for write
    """

    def _config(self):
        self.DATA_WIDTH = Param(8)
        self.USE_IN = Param(True)
        self.USE_OUT = Param(True)

    def _declr(self):
        if self.USE_IN:
            self.din = VectSignal(self.DATA_WIDTH, masterDir=D.IN)
        if self.USE_OUT:
            with self._paramsShared():
                self.dout = VldSynced()

    def _initSimAgent(self, sim: HdlSimulator):
        self._ag = RegCntrlAgent(sim, self)
